/*  -----------------------------------------------
    |                  VERONA                     |
    |     VErtex Reconstruction ON Analysis       |
    |                version 1.0                  |
    |                                             |
    |       Package for Analysing ROVER           |
    |            Monte Carlo data                 |
    |                                             |
    |     Authors: Jeroen van Tilburg (UU)        |
    |              Thomas Perschke (MPI-Munich)   |
    -----------------------------------------------  */

#include "../cats/include/rover.hh"
#include "../cats/include/rovermatrix.h"
#include "CLHEP/Hist/HBookTuple.h"
 
HBookTuple vertexntup("Reconstructed vertex data ntuple",100);
int evno=0;

static int OUTPUT  = 0;    //  Line output      (0=no, 1=yes)
static int HSTO    = 0;    //  Histogram output (0=no, 1=yes)
static int MC      = 1;    //  Monte Carlo data (0=no, 1=yes)
static int USERSEG = 0;    //  Use RSEG tracks  (0=no, 1=yes)
static int USERTRA = 0;    //  Use RTRA tracks  (0=no, 1=yes)

static float cut_x = 0.0300; //  Cut to determine ghost/missed vertices
static float cut_y = 0.0300; //  idem
static float cut_z = 0.2000; //  idem 
static int firstcall = 1;

//static int mintracksinvertex = 3;
     
static int MCvert  = 0;
static int MCfound = 0;
static int RCvert  = 0;
//static int RCfound = 0;

//===================================================================

void Lover::verona_init(){
  // Read parameters for VERONA
  
  OUTPUT  = 1;
  HSTO    = 1;
  USERSEG = 1;
  USERTRA = 0;
    

  if (ArteTable<MCIN>::booked()){
    MC = 1; // simu data
  }
  else{
    MC = 0; // real data
  }


  
  if(HSTO) start_hsto();
}

//===================================================================

void Lover::verona_event(){
  if (firstcall){
    verona_init();
    firstcall = 0;
  }
  if(HSTO) fill_track_hsto();
  if(HSTO) fill_ver_hsto();
}

//===================================================================

void Lover::start_hsto(void){

  // Online monitoring histo's
  hbook2(1000,"All tracks in target-plane z=-1.1", 500, -0.5, 0.5, 500, -0.5, 0.5);
  hbook2(1001,"All tracks in target-plane z=-5.0", 500, -0.5, 0.5, 500, -0.5, 0.5);
  
  // Errors in the vertex position
  hbook1(1121,"Rms x", 20, 0.0, 0.02);
  hbook1(1122,"Rms y", 20, 0.0, 0.02);
  hbook1(1123,"Rms z", 20, 0.0, 0.2);
  
  //  Resolution and pulls using MC
  hbook1(1101,"x_rec - x_mc", 50, -0.1, 0.1);
  hbook1(1102,"y_rec - y_mc", 50, -0.1, 0.1);
  hbook1(1103,"z_rec - z_mc", 50, -0.5, 0.5);
  hbook1(1111,"x_rec - x_mc pull", 30, -10.0, 10.0);
  hbook1(1112,"y_rec - y_mc pull", 30, -10.0, 10.0);
  hbook1(1113,"z_rec - z_mc pull", 30, -10.0, 10.0);
  
  // Target resolutions and pulls
  hbook1(1201,"x_rec - x_target", 50, -0.1, 0.1);
  hbook1(1202,"y_rec - y_target", 50, -0.1, 0.1);
  hbook1(1203,"z_rec - z_target", 50, -0.5, 0.5);
  hbook1(1211,"x_rec - x_target pull", 30, -10.0, 10.0);
  hbook1(1212,"y_rec - y_target pull", 30, -10.0, 10.0);
  hbook1(1213,"z_rec - z_target pull", 30, -10.0, 10.0);
  
  // Vertices
  hbook1(1600,"Num of Rec. Prim. vertices per event", 11, -0.5, 10.5);
  hbook1(1610,"Num of tracks in vertex", 26, -0.5, 25.5);
  hbook2(1620,"Wire correlation", 15, 0.75, 8.25, 15, 0.25, 8.25);
  hbook1(1630,"Ghost rate", 2, -0.5, 1.5);
  hbook1(1640,"Miss rate", 2, -0.5, 1.5);
  hbook1(1641,"Reco efficiency", 100, 0.0, 20.0);
}


//===================================================================

void Lover::fill_ver_hsto(void){

  cout.setf(ios::fixed);
  cout.setf(ios::showpoint);
  cout.precision(4);

  evno++;
  vector_m mc_vert_pos;
  // 
  // Read the MC vertices and put them in mc_vert_pos
  // and count the missed vertices
  // 
  if( MC ){
    int NumberOfPrimMC = 0;
    int NumberOfPrimMCfound = 0;
    for( ArteTable<MVER>::iterator mverIt = ArteTable<MVER>::begin();
	 mverIt != ArteTable<MVER>::end();++mverIt){
      vector<double> pos(3,0.);
      if( (*mverIt)->flag == 1){

	int mtrafound = 0;
	for (ArteTable<MTRA>::iterator mtraIt = ArteTable<MTRA>::begin();
	     mtraIt != ArteTable<MTRA>::end(); ++mtraIt){
	  if ((*mtraIt)->bver != (*mverIt)) continue;
	  int nhitb = 0;
	  for ( ArteRelationIterator<MIMP> mimpIt = (*mtraIt)->mimp.begin();
		mimpIt != (*mtraIt)->mimp.end(); ++mimpIt){
	    if ((*mimpIt)->gede->cmp != 1) continue;
	    nhitb += distance((*mimpIt)->hitb.begin(),(*mimpIt)->hitb.end()); 
	  }
	  if (nhitb>=5){
	    //	    cout<<(*mverIt)<<" found mtra "<<(*mtraIt)<<" with "
	    //		<<nhitb<<" hitb"<<endl;
	    mtrafound++;
	  }
	}
	cout<<"Vertex with "<<mtrafound<<" tracks"<<endl;

	if (mtrafound<mintracksinvertex) continue;

	NumberOfPrimMC++;
	pos[0] = double((*mverIt)->x);
	pos[1] = double((*mverIt)->y);
	pos[2] = double((*mverIt)->z);
	mc_vert_pos[NumberOfPrimMC] = pos;
	int miss = 1;
	for( ArteTable<RVER>::iterator rverIt = ArteTable<RVER>::begin();
	     rverIt != ArteTable<RVER>::end() ; ++rverIt){
	  if( fabs(pos[0] - (*rverIt)->xv) < cut_x &&
	      fabs(pos[1] - (*rverIt)->yv) < cut_y &&
	      fabs(pos[2] - (*rverIt)->zv) < cut_z ){
	    miss = 0 ;
	    cout<<(*mverIt)<<" OK  ";
	  }
	  else{
	    cout<<(*mverIt)<<" NOT ";
	  }
	  cout<<"distance   "
	      <<fabs(pos[0]-(*rverIt)->xv)<<" "
	      <<fabs(pos[1]-(*rverIt)->yv)<<" "
	      <<fabs(pos[2]-(*rverIt)->zv)<<endl;
	  cout<<"        covariance "
	      <<sqrt((*rverIt)->cv[0])<<" "
	      <<sqrt((*rverIt)->cv[2])<<" "
	      <<sqrt((*rverIt)->cv[5])<<" ==> "
	      <<(*rverIt)->nt<<endl;
	}
	hf1(1640, float(miss), 1.0);
	hf1(1641, mtrafound+0.4*miss, 1.0);
	MCvert++;
	if (miss==0){
	  MCfound++;
	  NumberOfPrimMCfound++;
	}
      }
    }

    RCvert+=distance(ArteTable<RVER>::begin(),ArteTable<RVER>::end());

    cout<<endl;
    cout<<"mc  simu  "<<NumberOfPrimMC<<endl;
    cout<<"mc  reco  "<<NumberOfPrimMCfound<<endl;
    cout<<"all reco  "<<distance(ArteTable<RVER>::begin(),ArteTable<RVER>::end())<<endl;
    cout<<endl;
    cout<<"MC  simu  "<<MCvert<<endl;
    cout<<"MC  reco  "<<MCfound<<endl;
    cout<<"ALL reco  "<<RCvert<<endl;
    if (MCvert != 0)
      cout<<"MC  effi  "<<float(MCfound)/float(MCvert)<<endl;
    if (RCvert != 0)
      cout<<"RC  ghost "<<float(RCvert-MCfound)/float(RCvert)<<endl;
    cout<<endl;
  }
  
  int NumberOfPrimRec = 0;
  int wire_old=0;
  for( ArteTable<RVER>::iterator rverIt = ArteTable<RVER>::begin();
       rverIt != ArteTable<RVER>::end() ; ++rverIt){
    NumberOfPrimRec++;
    vector<double> pos(3,0.);
    vector<double> cov(9,0.);
    int horver, wire;
    float wire_pos, wire_z;
    vector<double> mcpos(3,0.);
    int mc_no = 0;          // associated MC vertex number
    int ghost = 0;          // 0=associated with MC vertex, 1=ghost
    
    pos[0] = (*rverIt)->xv;
    pos[1] = (*rverIt)->yv;
    pos[2] = (*rverIt)->zv;
    cov[0] = (*rverIt)->cv[0];      cov[1] = (*rverIt)->cv[1];
    cov[2] = (*rverIt)->cv[3];      cov[3] = (*rverIt)->cv[1];
    cov[4] = (*rverIt)->cv[2];      cov[5] = (*rverIt)->cv[4];
    cov[6] = (*rverIt)->cv[3];      cov[7] = (*rverIt)->cv[4];
    cov[8] = (*rverIt)->cv[5];
    
    hf1(1121, sqrt(cov[0]), 1.0);
    hf1(1122, sqrt(cov[4]), 1.0);
    hf1(1123, sqrt(cov[8]), 1.0);
    hf1(1610, (*rverIt)->nt, 1.0);
    
    vertexntup.column("evno", evno);
    vertexntup.column("xrec", pos[0]);
    vertexntup.column("yrec", pos[1]);
    vertexntup.column("zrec", pos[2]);
    vertexntup.column("xsig", sqrt(cov[0]));
    vertexntup.column("ysig", sqrt(cov[4]));
    vertexntup.column("zsig", sqrt(cov[8]));
    vertexntup.column("nt", (*rverIt)->nt);
    //
    // Fill target resolutions
    //
    if( closestwire(pos, cov, horver, wire_pos, wire_z, wire) ){
      vertexntup.column("wire", wire) ;
      if( horver == 0 ){  
	hf1(1202, pos[1] - wire_pos, 1.0);
	hf1(1212, (pos[1] - wire_pos)/sqrt(cov[4]), 1.0);
	vertexntup.column("tarresxy", pos[1] - wire_pos) ;
      }
      if( horver == 1 ){ 
	hf1(1201, pos[0] - wire_pos, 1.0);
	hf1(1211, (pos[0] - wire_pos)/sqrt(cov[0]), 1.0);
	vertexntup.column("tarresxy", pos[0] - wire_pos) ;
      }
      hf1(1203, pos[2] - wire_z, 1.0);
      hf1(1213, (pos[2] - wire_z)/sqrt(cov[8]), 1.0);
      vertexntup.column("wirepos", wire_pos) ;
      vertexntup.column("wirez", wire_z); 
      
      if( NumberOfPrimRec > 1 ) hfill(1620, wire_old, wire, 1.0);
      wire_old = wire;
    }
    //
    // Fill MC resolutions
    //
    if( MC != 1 ) {
      vertexntup.dumpData();
      continue;
    }
    if(getclosestvertex(pos,cov,mc_vert_pos,mc_no) != 0){
      vertexntup.dumpData();
      continue;
    }
    mcpos = mc_vert_pos[mc_no];
    
    if( fabs(pos[0]-mcpos[0]) > cut_x ||
	fabs(pos[1]-mcpos[1]) > cut_y ||
	fabs(pos[2]-mcpos[2]) > cut_z )  ghost=1;  // Count ghost rate
    
    hf1(1630, float(ghost), 1.0);
    hf1(1101,(pos[0]-mcpos[0]),1.0);
    hf1(1102,(pos[1]-mcpos[1]),1.0);
    hf1(1103,(pos[2]-mcpos[2]),1.0);
    hf1(1111,( (pos[0]-mcpos[0]) / sqrt(cov[0]) ),1.0);
    hf1(1112,( (pos[1]-mcpos[1]) / sqrt(cov[4]) ),1.0);
    hf1(1113,( (pos[2]-mcpos[2]) / sqrt(cov[8]) ),1.0);
    
    vertexntup.column("xmc", mcpos[0]);
    vertexntup.column("ymc", mcpos[1]);
    vertexntup.column("zmc", mcpos[2]);
    vertexntup.column("ghost", ghost);
    vertexntup.dumpData();
    
  } // End of loop over RVER

  hf1(1600, float(NumberOfPrimRec), 1.0);
}


//===================================================================

void Lover::fill_track_hsto(void){
  float x, y;
  
  
  if( USERSEG ){
    for( ArteTable<RSEG>::iterator rsegIt = ArteTable<RSEG>::begin();
	 rsegIt != ArteTable<RSEG>::end(); ++rsegIt ){
      x = (*rsegIt)->xf + (-1.1 - (*rsegIt)->zf)*(*rsegIt)->txf;
      y = (*rsegIt)->yf + (-1.1 - (*rsegIt)->zf)*(*rsegIt)->tyf;
      hfill(1000,x,y,1.0);
      x = (*rsegIt)->xf + (-5.0 - (*rsegIt)->zf)*(*rsegIt)->txf;
      y = (*rsegIt)->yf + (-5.0 - (*rsegIt)->zf)*(*rsegIt)->tyf;
      hfill(1001,x,y,1.0);
    }
  }
  
  if( USERTRA ) {
    for( ArteTable<RTRA>::iterator rtraIt = ArteTable<RTRA>::begin();
	 rtraIt != ArteTable<RTRA>::end(); ++rtraIt ){
      if( ((*rtraIt)->hsi) > 7 ){
	x = (*rtraIt)->xf + (-1.1 - (*rtraIt)->zf)*(*rtraIt)->txf;
	y = (*rtraIt)->yf + (-1.1 - (*rtraIt)->zf)*(*rtraIt)->tyf;
	hfill(1000,x,y,1.0);
	x = (*rtraIt)->xf + (-5.0 - (*rtraIt)->zf)*(*rtraIt)->txf;
	y = (*rtraIt)->yf + (-5.0 - (*rtraIt)->zf)*(*rtraIt)->tyf;
	hfill(1001, x, y, 1.0);
      }
    }
    
  } // if (RTRA)
}

       
//===================================================================

int Lover::getclosestvertex(vector<double> pos,vector<double> cov,vector_m
                     mcvertexmap,int& mc_no){
  
  int best=0;
  double min = 1.0e8;
  
  if(mtxinv(cov,3)){
    cout << "verona::getclosestvertex() matrixinversion failed" << endl;
    return 1;
  }
 
  for(vector_mi itmc = mcvertexmap.begin();itmc != mcvertexmap.end();++itmc){
    vector<double> cx(3,0.);
    
    for(int i=0;i<3;++i)
      for(int j=0;j<3;++j)
	cx[i] += cov[i+3*j]*((itmc->second)[j]-(pos)[j]);
 
    double chi2 = 0.;
    for(int i=0;i<3;++i)
      chi2 += ((itmc->second)[i]-(pos)[i])*cx[i];
 
    if(chi2<min){
      min = chi2;
      best = itmc->first;
    }    
  }
      
  if(best == 0){
    cout << "verona::getclosestvertex() no best track found" << endl;
    return 1;
  } 
  
  mc_no = best;
  
  return 0;
  
}

//===================================================================

int Lover::closestwire(vector<double> pos,vector<double> cov,int& horver, 
		       float& wire_pos, float& wire_z, int& wire){
  double min = 1.0e8;
  double dist;
  int hv;
  wire = -1;
  
  for(Target_mi tarit = targets.begin(); tarit != targets.end(); 
      ++tarit){
    

    // transform to global coordinates !!!
    
    float local[]  = {0,0,0,0,0};
    float global[] = {0,0,0,0,0};
    
    local[0] = (tarit->second).x;
    local[1] = (tarit->second).y;
    local[2] = (tarit->second).z;
    acLocalToGlobal(local, global, 1);
    float xt = global[0];
    float yt = global[1];
    float zt = global[2];

    hv = (tarit->second).flaghorver;

    if( hv == 0 ){
      dist = (yt - pos[1])*(yt - pos[1])/cov[4] +
	(zt - pos[2])*(zt - pos[2])/cov[8];
    }
    else{
      dist = (xt - pos[0])*(xt - pos[0])/cov[0] +
	(zt - pos[2])*(zt - pos[2])/cov[8];
    }
    if( dist < min){
      if( hv == 0) wire_pos = yt ;
      if( hv == 1) wire_pos = xt ;
      wire_z = zt ;
      wire = (tarit->second).wnr;
      horver = hv ;
      min = dist;
    }
  }
  
  if( min == 1.0e8 ){
    cout << "verona::closestwire() no best wire found" << endl;
    return 0;
  } 
    
  return 1;
  
}

//===================================================================
