
// *****************************************************************************
//
// source:    faEmg.C
//
// created:   30 Jul 1997
//
// Author:    Andreas Gellrich
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: Andreas.Gellrich@desy.de
//
// changes:         
// 12 Jan 1999 (AG) removed emg status flag
// 14 Jan 1999 (AG) polishing
// 02 Feb 1999 (AG) polishing
// 22 Feb 1999 (AG) no READALL in emg for cycle mode
// 25 Mar 1999 (AG) correct emg_consum() and emg_get_event()
// 01 Jun 1999 (AG) put includes
// 09 Sep 1999 (AG) always use cout
// 03 Nov 1999 (AG) renamed CYCLE to NOBLOCK
// 23 Nov 1999 (AG) do not wait for emg
// 25 Nov 1999 (AG) introduce SMC
// 08 Dec 1999 (AG) count events
// 10 Dec 1999 (AG) event counting is done by Arte
// 24 Jan 2000 (AG) correct usleep()
// 29 Feb 2000 (AG) change sleep time to 50 msec
// 22 Mar 2000 (AG) polish emg
// 23 Mar 2000 (AG) polish emg
//
// *****************************************************************************

// standard includes
#include <iostream.h>
#include <unistd.h>

// ONLINE includes
#include "rpm.h"
#include "smc.h"
#include "rhp.h"
#include "emg.h"
#include "grr.h"
#include "par.h"

// Arte includes
#include "farm/fa.hh"
#include "farm/farm.hh"

// externals
extern int  FADEB;     // debug flag
extern int  NOBLOCK;   // no block emg
extern int  RPMINIT;   // init rpm/rhp
extern char *PARNAME;  // emg partition name;
extern int  SEG_NUM;   // emg partitions
extern int  SEG_LEN;   // emg segment size

extern char *evtBuf;   // event buffer

// constants
//const int MUSEC_SLEEP = 500000;
const int MUSEC_SLEEP = 50000;

// prototypes
extern "C" void atstopcc_( char *txt );
#ifdef __linux
extern "C" void usleep( unsigned long musec );
#endif

// *****************************************************************************
//
// name:      faReadEmg
//
// purpose:   read event from emg
//
// called by: faReadEmg
//
// input:     void
// return:    int  rc
//
// *****************************************************************************

int
faReadEmg()
{
  // pre-sets
  int        rc    = GRR_OK;
  static int count = 0;
  
  // print
  if( count < 10 ){
#ifdef _OUT_
    cout << "EMG: Waiting for event ....." << endl;
#endif
  }

  //
  // get event
  //
  //int flags = 0;
  int flags  = NOWAIT;
  while( 1 ){
    if( ( rc = emg_get_event( (int *) evtBuf, SEG_LEN, flags ) ) == GRR_OK ){
      if( count < 10 ){
	count++;
#ifdef _OUT_
	cout << "EMG: Got event" << endl;
#endif
      }
      // check for rhp messages
      if( RPMINIT ){

	if( FADEB ){
	  cout << "EMG: Check for rpm messages." << endl;
	}

	// complete transition
	if( smc_if_trans_pend() == SMC_OK ) smc_complete_trans();

	// dispatch
	while( rhp_wait_for_one_msg_and_dispatch( NO_BLOCK ) == GRR_OK );
      }
      break;
    }
    else if( rc == GRR_EMPTY ){
#ifdef _OUT_
      cout << "EMG: Queue empty. Sleep " << MUSEC_SLEEP << " musec and retry." << endl;
#endif
      // check for rhp messages
      if( RPMINIT ){

	if( FADEB ){
	  cout << "EMG: Check for rpm messages." << endl;
	}

	// complete transition
	if( smc_if_trans_pend() == SMC_OK ) smc_complete_trans();

	// dispatch
	while( rhp_wait_for_one_msg_and_dispatch( NO_BLOCK ) == GRR_OK );
      }

      // sleep
      usleep( MUSEC_SLEEP );
    }
    else {
#ifdef _ERR_
      cout << "EMG: ERROR: getting event rc=" << rc << endl;
#endif
      break;
    }
  }
  return rc;
}

// *****************************************************************************
//
// name:      faInitEmg
//
// purpose:   initialize emg
//
// called by: fainit_
//
// input:     void
// return:    int  error code
//
// *****************************************************************************

int
faInitEmg()
{
  //
  // flags
  //
  int consumFlags = READALL;
  if( NOBLOCK ){
    consumFlags = 0;
  }

  int attachFlags = 0;       // create OR use
  //int attachFlags = EMG_OLD; // require existing emg

  //
  // init emg
  //
  int rc = GRR_OK;
  if( ( rc = emg_init() ) != GRR_OK ){ // initialize emg
#ifdef _OUT_
    cout << "EMG: ERROR: init rc=" << rc << endl;
#endif
    return rc;
  }

  //
  // emg ctrl
  //
#ifdef _OUT_
  cout << "EMG: "
       << "partition=" << PARNAME     << " "
       << "segments="  << SEG_NUM     << " "
       << "length="    << SEG_LEN     << " "
       << "attach="    << consumFlags << " "
       << "consum="    << attachFlags << endl;
#endif

  //
  // SEG_NUM
  //
  if( ( rc = emg_ctrl( NSEGS, &SEG_NUM ) ) != GRR_OK ){ // set parameters
#ifdef _OUT_
    cout << "EMG: ERROR: segments rc=" << rc << endl;
#endif
    return rc;
  }

  //
  // SEG_LEN
  //
  if( ( rc = emg_ctrl( SEGLEN, &SEG_LEN ) ) != GRR_OK ){
#ifdef _OUT_
    cout << "EMG: ERROR: length rc=" << rc << endl;
#endif
    return rc;
  }

  //
  // attach to emg
  //
  if( ( rc = emg_attach( PARNAME, attachFlags ) ) == GRR_FISHY ){
#ifdef _OUT_
    cout << "EMG: ERROR: (GRR_FISHY) partition NOT attached." << endl;
#endif
    ARTE_error( GRR_SFATAL, "EMG: ERROR: (GRR_FISHY)", rc );
    atstopcc_( "EMG: ERROR: (GRR_FISHY) partition NOT attached." );
    //return rc;
  }
  else if( rc == GRR_UNKNOWN ){
#ifdef _OUT_
    cout << "EMG: ERROR: (GRR_UNKNOWN) partition NOT found." << endl;
#endif
    ARTE_error( GRR_SFATAL, "EMG: ERROR: (GRR_UNKNOWN)", rc );
    atstopcc_( "EMG: ERROR: (GRR_UNKNOWN) partition NOT found." );
    //return rc;
  }
  else if( rc == GRR_INCOMPATIBLE ){
#ifdef _OUT_
    cout << "EMG: ERROR: (GRR_INCOMPATIBLE) partition USED somewhere else?" << endl;
#endif
    ARTE_error( GRR_SFATAL, "EMG: ERROR: (GRR_INCOMPATIBLE)", rc );
    atstopcc_( "EMG: ERROR: (GRR_INCOMPATIBLE) partition USED somewhere else?" );
    //return rc;
  }
  else if( rc != GRR_OK ){
#ifdef _OUT_
    cout << "EMG: ERROR: rc=" << rc << endl;
#endif
    ARTE_error( GRR_SFATAL, "EMG: ERROR", rc );
    atstopcc_( "EMG: ERROR" );
    //return rc;
  }

  //
  // consum
  //
  if( ( rc = emg_consum( -1, -1, 1, consumFlags ) ) != GRR_OK ){ // consum emg
#ifdef _OUT_
    cout << "EMG: ERROR: consuming "
	 << "consumFlags=" << consumFlags << " "
	 << "rc="    << rc    << endl;
#endif
    return rc; // error
  }
  
  return rc; // okay
}

// *****************************************************************************
//
// name:      faStopEmg
//
// purpose:   terminate
//
// called by: Fortran program ATSTOP
//
// input:     void
// return:    int  error code
//                 =0: ok
//                 <0: error
//
// *****************************************************************************

int
faStopEmg()
{
  int rc = GRR_OK;
  if( ( rc = emg_release( PARNAME, 0 ) ) == GRR_OK ){ // realease emg
#ifdef _OUT_
    cout << "EMG: released." << endl;
#endif
    return rc;
  }
  else {
#ifdef _ERR_
    cout << "EMG: ERROR: releasing rc=" << rc << endl;
#endif
    return rc;
  }
}
