/***********************************************************************
*-- Author :    Siegmund Nowak
*
*  This is the CERN user routine called at the end of each tracking step
*  Here actions are collected which are specific for HERA-B but general
*  for all HERA-B users. The individual user routine is called USSTEP 
*  The actions called from GUSTEP are:
*  - drawing of tracks
*  - filling of temporary MIMP/MCAL banks for sensitive detectors
*  - storing of new generated particles, restoring of STAK particles 
*  - calling of USSTEP
*  - limiting of spiraling particles in the magnet
*  - setting length and flag parameters for ending tracks
*  - limiting of endless loops on boundaries for flat tracks 
***********************************************************************/
#include <string.h>
#include <math.h>
#include <vector.h>
#include <stdio.h>
#include <ctype.h>
#include "gean/gckine.h"
#include "gean/gcking.h"
#include "gean/gcflag.h"
#include "gean/gctrak.h"
#include "gean/gcvolu.h"
#include "gean/gctmed.h"
#include "gean/gcsets.h"
#include "gean/gcnum.h"
#include "gean/gcpush.h"
#include "gean/ZGMOTH.h"
#include "gean/ZGESL.h"
#include "gean/ZGDETV.h"
#include "gean/zgkine.hh"
#include "gean/ZGTRDE.h"
#include "arte/arte.hh"
#include "arte/ArteFlag.h"
#include "arte/ArteUtil.h"
#include "arte/ArteMessage.hh"

extern "C" {
   void gustep_();               
   void mustcheck_(char*,int*);
   void gsxyz_();                
   void gdcxyz_();
   void trade_();
   void gsking_(int *);          
   void ghstak_();
   void muturb_();               
   void usstep_();
}
   void crhibk(int);            
   void gspiral();

void gustep_() {
   extern vector<hgmtra> *hg_mtra;
   char cname[4];

// define some useful quantities
// select volumina where particles are only in stack (jspec!= )  
   const int maxspec = 9;   
   char idnam[maxspec*4+1]; 
   strcpy(idnam,"CAY1CAY2CAY3CASCCAPBCATUSCRPPOLSPOLC");
//               ^   ^   ^   ^   ^   ^   ^   ^   ^
   memcpy(cname,&gcvolu_.names[gcvolu_.nlevel-1],4);
   for(int i =0; i<4; i++) cname[i] = toupper(cname[i]);  
   int jspec = 0;    int icheck = 0;
   for(int i=0; i < 4*maxspec; i+=4){
      if (!strncmp(cname,&idnam[i],4)) {
         jspec = i/4 + 1;  break; 
      }
   }  
// speciality for muon system
   char mname[5]; 
   strncpy(mname,cname,4);
   mname[4] = '\0';
   mustcheck_(mname,&icheck);
   if(icheck != 0) jspec =maxspec+1;

// get some track valuestrack values
   float  abs_charge = fabs(gckine_.charge); 
   vector<hgmtra>::iterator mt = hg_mtra->begin() + gckine_.itra -1;

// draw tracks if wanted
   if(!arteflag_.ibatch && gcflag_.iswit[0] && abs_charge >= 1.e-2) {
       gsxyz_();   gdcxyz_();
   }

   int istpu = 0;
   if(jspec > 0 && jspec <= 6) { 
//    make a calorimeter endpoint mtcp  at the entry of the calorimeter 
      if(!gckine_.istak && !mt->ep ) crhibk(5);
   }    

   else if(!strncmp(cname,"KILL",4)) {  
//      stop the particle if it enters in volume "kill"     
        gctrak_.istop=1;     goto L60;
   }

   else if(!strncmp(cname,"TR",2) || gckine_.ipart == 48 ) { 
//      call special code for trd-simulation
        if(trdsw_.itrswt == 1) trade_(); 
   }

   else if(!strncmp(cname,"MU",2)) {
//      switch on fast electron tracing by means of mututb
        extern int muon_turbfl;
        if(muon_turbfl == 1) muturb_();
   }

   else if(gctmed_.ifield &&  abs_charge >= .1 
        && gctrak_.vect[6] < 1. && !mt->nhit){
//      stop tracks to avoid spirals in magnetic field
        gspiral();
        if(gctrak_.istop && !gckine_.istak && arteflag_.mimpsw) istpu = -1; 
   }

// avoid long loops for flat tracks on boundaries (temporary)
   if(fabs(gctrak_.vect[5]) <= .04) {
      static int nust,nstold,itrold,istold,hbname,hbnum,vflag; 
      if (gctrak_.nstep == 0) nust = 0;
      if(gctrak_.inwvol == 2 && gctrak_.istop == 0) { 
         hbname = gcvolu_.names[gcvolu_.nlevel-1];
         hbnum  = gcvolu_.number[gcvolu_.nlevel-1];
         itrold = gckine_.itra;
         istold = gckine_.istak;
         nstold = gctrak_.nstep;
         vflag = 1;
      }
      if( gctrak_.inwvol == 1 && vflag == 1 &&  gctrak_.nstep >= nstold
          && gckine_.itra == itrold && gckine_.istak == istold ) { 
         if( hbname == gcvolu_.names[ gcvolu_.nlevel-1] &&
             hbnum  == gcvolu_.number[gcvolu_.nlevel-1]) {
             nust = nust++;
             if(nust >= 20) {
                gctrak_.vect[0] = gctrak_.vect[0] + .01 *gctrak_.vect[3]; 
                gctrak_.vect[1] = gctrak_.vect[1] + .01 *gctrak_.vect[4]; 
                gctrak_.vect[2] = gctrak_.vect[2] + .01 *gctrak_.vect[5]; 
             }
             if(nust >=30){ 
	       //                printf("problem %.04s %d %f\n",cname,gckine_.ipart,gctrak_.vect[5]);
                gctrak_.istop = 1;
             }
         }
         else { 
	   //	   if(nust >=20 ) printf("nust %d \n",nust);
            nust = 0;
	 }
         vflag=0;
      }  
   } 

// mimps will be written out
   if(arteflag_.mimpsw) {
      if(!gcsets_.idtype) {}
      else if(gcsets_.idtype == LGESL_SIL && zgdetv_.DETV[LGESL_SIL-1] == 'S'){ 
             if(gctrak_.inwvol == 1 && abs_charge > 1.e-2)  crhibk(6);
      }   
      else if(gcsets_.idtype == LGESL_OTR && zgdetv_.DETV[LGESL_OTR-1] == 'L'){
             if(gctrak_.inwvol == 1 && abs_charge > 1.e-2) crhibk(8);
      }
      else if(gcsets_.idtype == LGESL_MSG && zgdetv_.DETV[LGESL_MSG-1] == 'S'){
             if(gctrak_.inwvol == 1 && abs_charge > 1.e-2) crhibk(6);
      }   
      else if(gcsets_.idtype == LGESL_ECA) {
             if(gctrak_.destep >0. ) crhibk(2);
      } 
      else if(gcsets_.idtype == LGESL_TRD){
             if(gctrak_.inwvol == 1 && abs_charge > 1.e-2 )  crhibk(7);
      }
      else if(gcsets_.idtype == LGESL_RIC) { 
             if(gctrak_.inwvol == 1 && abs_charge > 1.e-2 )  crhibk(1);
      }
      else if(gcsets_.idtype == LGESL_HOD1 || gcsets_.idtype == LGESL_HOD2 
             || gcsets_.idtype == LGESL_HOD3) {
             if(gctrak_.destep>0) crhibk(3);
      }
      else { if(gctrak_.inwvol == 1 && abs_charge > 1.e-2 )  crhibk(1);
      }
      if(gcsets_.idtype == LGESL_TRD && trdsw_.itrswt == 1 
                                     && gctrak_.destep > 0. ){
	if(abs_charge > 1.e-2 || gckine_.ipart == 48) crhibk(4);
      }
   }
      
// store newly generated particles in jkine/jvertx if they are outside
// the volumes idnam and in stak only if they are inside
   if(gcking_.ngkine > 0) {
      int ifl0 = 1;   if(jspec != 0) ifl0 = 0; 
      if(ifl0 == 1 && 
               (gcnum_.ntrack + gcpush_.npkine > 64000 
             || gcnum_.ntrack + gcking_.ngkine > 64000)) {
	 if(gcflag_.ieotri !=0) {  
	    arteMsg << "event skipped because of more than 64 k geant tracks";
            arteMsg.warn();
         }
         gcflag_.ieotri = 1;
         arteutil_.eventout = 0;
         return;
      } 
      for (int i=0; i<gcking_.ngkine; i++)  gcking_.iflgk[i] = ifl0;  
      int i0 = 0;     gsking_(&i0);  
   }  

// restore stak particles in jkine/jvertx banks again, if they come
// out of volumes idnam
   if(gckine_.istak != 0 && gctrak_.inwvol == 1) {
      if(!strncmp(cname,"TR",2) || gckine_.ipart == 48) {}
      else if(jspec == 0) { 
        if(gcnum_.ntrack + gcpush_.nckine > 64000 ) {
	   if(gcflag_.ieotri !=0) {
	      arteMsg << "event skipped because of more than 64 k geant tracks";
              arteMsg.warn();
           } 
           gcflag_.ieotri = 1;
           arteutil_.eventout = 0;
           return;
        }
        ghstak_();  gctrak_.istop = 1; 
      } 
   }



L60:   // call user subroutine
    if(gctrak_.istop && !istpu ) istpu = 1;
    usstep_();


    if(gctrak_.istop) {               // stopping tracks: give length and flag  
       if(!istpu) istpu = 2;

       if(!gckine_.istak && arteflag_.mimpsw) {
          mt->len = gctrak_.sleng;

          if(gctrak_.inwvol == 3) {                  // leaving the detector  
             mt->flag = 5;   return; 
          }
          if(istpu < 0) {                       // spiraling cut
	     mt->flag = 7;   return;
          }
          if(istpu == 2){                       // user stop  
             mt->flag = 8;   return;
          }
          if(!strncmp(cname,"KILL",4)) {        // volume KILL
             mt->flag = 5;    return; 
          }

          for(int i=0; i<gctrak_.nmec; i++) {

             if(gctrak_.lmec[i] == 5) {         // decay
                mt->flag = 3;   return; 
             }

             if((gctrak_.lmec[i] >= 6 && gctrak_.lmec[i]<= 8) 
                   ||   gctrak_.lmec[i] ==11 || gctrak_.lmec[i]== 12) {
                mt->flag = 4;   return;         //  interaction
              }

             if(gctrak_.lmec[i] == 30) {        // energy cutoff
                 mt->flag = 6;   return; 
             }
          }
       }
    }
    
    return;

       
}



