#ifndef _RCMATCHBUILDER_HH_
#define _RCMATCHBUILDER_HH_
//---------------------------------------------------------------------
// Created: 13.10.97,   Olya Igonkina

// MARPLE builder class of Points for Reconstruction

// !!! important
// BE SURE that the classes used for matching are derived from RcPoint class
// parameters of match will not deleted

//---------------------------------------------------------------------


#include <vector>
#include <iterator>
#include <math.h>

#include "arte/arte.hh"

#include "marple/RcPoint.hh"
#include "marple/RcMatchPoint.hh"
#include "marple/RcMatchConst.h"

/** {\bf MARPLE main class to do match }                        \\

    The procedure how to do match is described in section \Ref{Usage} of
    this manual.

    In addition the following functions present the result of matching:     \\
    vectorLeft().size()  - number of RcPoints to be matched from left side; \\
    vectorRight().size() - number of RcPoints to be matched from right side; \\
    NrealPair()          - number of pairs of RcPoints saved in left and right
                           vectors corresponded to one MTRA particle;       \\
    NrefPair()           - number of pairs of RcPoints saved in left and right
                           vectors corresponded to one reference MTRA particle;\\
    matchList().size()   - number of matched pairs;                          \\
    Nbad()               - number of matched pairs of RcPoints corresponded 
                           to the different MTRA particles;                  \\
    Nlost()              - number of lost pairs of RcPoints    
                           correspond to the one MTRA particle \\
			   with momentum > 1 GeV/c and created at z < 100 cm;\\

   Two {\bf important} points should be kept in mind:
   \begin{itemize}
   \item The class RcMatchBuilder is a template one. It means that 
   to use the object of RcMatchBuilder type the two classes  
   {\bf PointsLeft} and {\bf PointsRight} should be defined.
   For correct usage of RcMatchBuilder they should be inherited 
   from the \Ref{RcPoint} class. The {\it mark} member field of RcPoint class 
   is used to manage the number of combination with the given object.
   \item To avoid garbage creation the cleaning functions should be 
   called in time. The best way is to invoke {\bf cleanAll()} function when 
   the object is not needed anymore. Do it always at the end of the
   logical slope of program. 
   \end{itemize}

   @memo MARPLE main class to do match 
*/
template <class PointsLeft, class PointsRight>
class RcMatchBuilder
{
public:
  ///              default constructor
  inline RcMatchBuilder(); 

  ///              make match of two vectors with rcConst match parameters
  RcMatchBuilder( std::vector<PointsLeft *> vecLeft,
		  std::vector<PointsRight *> vecRight,
		  RcMatchConst & rcConst);

  ///               destructor
  inline ~RcMatchBuilder();

  ///               put left vector of RcPoints to be matched
  inline void addLeft(  std::vector<PointsLeft *> & vecLeft);
  ///               put right vector of RcPoints to be matched
  inline void addRight( std::vector<PointsRight *> & vecRight);

  ///               add RcPoint to left vector to be matched
  inline void addPointLeft(  PointsLeft & pointLeft);
  ///               add RcPoint to right vector to be matched
  inline void addPointRight( PointsRight & pointRight);

  ///               set parameters of match see \Ref{RcMatchConst}
  inline void setConst( RcMatchConst & rcConst)
    { fptRcConst = &rcConst; }; 

  ///               return left vector to be matched
  inline std::vector< PointsLeft  *> vectorLeft()  const
    { return fvecLeft;   };
  ///               return right vector to be matched
  inline std::vector< PointsRight *> vectorRight() const
    { return fvecRight;  };

  ///               return left origin of matchList()[iMatch],  0 <= iMatch < matchList().size() 
  inline PointsLeft * leftOrigin( unsigned int iMatch ) const;

  ///               return right  origin of matchList()[iMatch],  0 <= iMatch < matchList().size() 
  inline PointsRight * rightOrigin( unsigned int iMatch ) const;
  
  ///       return vector of found RcMatchPoints
  inline std::vector< RcMatchPoint *> matchList()   const
    { return fmatchList; };
  ///               return parameters of match
  inline RcMatchConst  * ptToMatchConst() const { return fptRcConst; };

  /// return real number of pairs of RcPoints correspond to one MTRA particle
  inline int    NrealPair()      const{ return fNrealPair; };
  /// return number of pairs of RcPoints correspond to one reference MTRA particle
  inline int    NrefPair()      const{ return fNfastReal; };
  ///               return number of fake pair
  inline int    Nbad()           const{ return fNbad;      }; 
  /// return the number of lost pairs of RcPoints correspond to one MTRA particle
  inline int    Nlost()          const{ return fNlost;     }; 

  //main operation :

  ///               make match between vectorLeft and vectorRight
  void doMatch();  
  ///               print information about matching
  void stat()   const; 
  ///               print information about matching, optional version
  void stat(int iLong )   const; 
  ///               fill ntuple \Ref{DELTAS} with cuts
  void fillDeltas(int iNtup = 0, int pulsNtup = 0) const; 
  ///               fill ntuple \Ref{MATCHNT} with match statistics
  void fillMatch( int iNtup = 0) const; 

  ///Kalman refit of left matched objects using mean parameters of RcMatchPoint
  void refitLeftFound(int & iErr);

  ///Kalman refit of left matched objects using mean parameters of RcMatchPoint
  void refitRightFound(int & iErr);


  void updateARTELeftFound(int & iErr) const;
  void updateARTERightFound(int & iErr) const;

  // save information into ARTE banks
  //void saveRTRA(int & iErr, int withRefit = 1) const;
  void saveRSEG(int & iErr ) const;
  //  void saveRCCL() const;
  /// save relation between RcPoints from left and right vectors which pass match procedure, return 0 in case of save errors 
  void saveRelation(int & iErr) const;

  //delete functions:  iStat = 1 delete statistics 

  //inline void deleteLeftPoint(  int iStat = 1);  // delete information
  //inline void deleteRightPoint(  int iStat = 1);  // delete information
  ///delete information saved in vectorLeft, if iStat = 1 statistics of match is also cleaned
  inline void deleteLeft(  int iStat = 1);  // delete information
  ///                 delete information saved in vectorRight
  inline void deleteRight( int iStat = 1);  
  ///                 delete match information 
  inline void deleteMatch( int iStat = 1);  
  /// delete all information, if iStat = 1 statistics of match is also cleaned
  inline void deleteAll(   int iStat = 1);
  /// delete all information ( the same as deleteAll() )
  inline void clearAll(   int iStat = 1);
  //                  clean statistics information
  inline void deleteStat();

  ///         set pointers of left vector to zero without touching of RcPoint 
  inline void eraseLeft(   int iStat = 1);
  ///         set pointers of right vector to zero without touching of RcPoint
  inline void eraseRight(  int iStat = 1);

  ///                 copy all information saved in RcMatchBuilder object
  inline RcMatchBuilder< PointsLeft, PointsRight> & operator = 
  (const RcMatchBuilder< PointsLeft, PointsRight> &rp );

protected:

  void saveRelation(const PointsLeft * orig1, 
		    const PointsRight * orig2, 
		    int & iErr) const;

  // save relation between original points and attached them to (match)segment
  void saveRelationWithMatch(const PointsLeft * orig1, 
			     const PointsRight * orig2, 
			     ArtePointer<RSEG> rseg,
			     int & iErr) const;

  ///                     left vector to be matched
  std::vector< PointsLeft   * >  fvecLeft;
  ///                     right vector to be matched
  std::vector< PointsRight  * >  fvecRight;
  ///                     vector with match information, filled by doMatch()
  std::vector< RcMatchPoint * >  fmatchList;
  ///                     constants used by matching
  RcMatchConst *            fptRcConst;

  ///   number of pairs of RcPoints from left and right vectors correspond to one MTRA particle
  int fNrealPair;
  ///   number of pairs of RcPoints from left and right vectors correspond to one reference MTRA particle
  int fNfastReal;
  ///           number of fake pairs found by matching    
  int fNbad;
  ///           number of lost reference pairs
  int fNlost;
};

//template <class PointsLeft, class PointsRight>
//ostream & operator << (ostream &, const RcMatchBuilder< PointsLeft, 
//		                                        PointsRight> &);

#include "marple/RcMatchBuilder.icc"

//unfortunately has to include that also to solve problem with template 
//#if __GNUC__ <= 2 && __GNUC_MINOR__ <= 7
#include "marple/RcMatchBuilder.C"
//#endif

#endif
