/* Defines FEDs for any Setup.
   Contains classes FED<AnyIChamber> 
                       = vector<Cables<AnyIChamber>::iterator>
   (see ptset/cables.hh).
   Thus FED is not a container of Cables. The holder of "Cable"s is
   the external structure "Cables" (note "s" in the end).
   FED is a container of Cables iterators. 
   Thus some cables can be in the "Cables" but not in the FEDs.

   Some FED slots can be not in use. 
   To find out whether there is a cable in the slot #n:
     bool is_connected(unsigned int n);
   (if there is no cable in the slot, this is marked internally by
   using the iterator == cables.end())
   
   Container of FED's:
   class FEDs<AnyIChamber> = map<unsigned int, FED<AnyIChamber> >

   FEDs and FED have 
     void dump(ostream& os);
   which prints the contents of the structures.

   Usually the programmer should use "Cables" and "FEDs".
   To connect them together one can use
     FEDs_Filler<AnyIChamber> 
   to construct the "interactive command" which fills the FEDs structure. 
   (See ptset/interactive.hh for details).

   Usage: some_menu["submenu"]("command_name") =
     make_interactive_command(FEDs_Filler<AnyIChamber>(feds, cables));
   One should supply "feds" and "cables" as parameters for the
   FEDs_Filler<AnyIChamber> constructor.
   When the "some_menu" is invoked
     some_menu(&is);
   the cables will be connected to the feds. 
   The format of the input stream should be:
     FED_number cable_1_name cable_2_name skip 5 cable_8_name ...
   Note the usage of the "skip 5".
   (I.e. for example 
   /submenu/command_name{ FED_number cable_1_name ... } )
   
   The FEDs_Filler<AnyIChamber> checks that the number of channels
   in cable is equal to that of in the FED slot and the total
   number of cables is equal to the number of slots in one FED.
   (Again, some slots can be "skipped").
   Therefore one should check global_error() (see global_error.hh)
   after some_menu(&is); call.

   There is an example in
     ptset/usage_of_cables_and_FEDs.example_readme
   */

// Author V.Balagura (balagura@mail.desy.de)   (21.02.99).

#ifndef _FED_hh
#define _FED_hh 1

#include "ptset/cables.hh"

template <class AnyIChamber>
class FEDs_Filler;

template <class AnyIChamber>
struct FED : public vector<Cables<AnyIChamber>::iterator> {
  static const unsigned int N_SLOTS            = 64;
  static const unsigned int N_CHANNELS_IN_SLOT = 16;

  bool is_connected(unsigned int n) { return (*this)[n] == dummy; }
  void dump(ostream& os);
private:
  Cables<AnyIChamber>::iterator dummy; // marks empty FED slots
  friend class FEDs_Filler<AnyIChamber>;
};

template <class AnyIChamber>
struct FEDs 
  : public map<unsigned int, FED<AnyIChamber>, less<unsigned int> > {
    void dump(ostream& os);
};

template <class AnyIChamber>
class FEDs_Filler {
public:
  FEDs_Filler(FEDs<AnyIChamber>* feds_container,
	      Cables<AnyIChamber>& cables_container) 
    : feds(*feds_container), cables(cables_container) {}

  void operator()(istream* is);
private:
  FEDs<AnyIChamber>& feds;
  Cables<AnyIChamber>& cables;

  void fill_cables(istream* is, FED<AnyIChamber>& fed);
  void set_readout_flag_for_cells(Cables<AnyIChamber>::iterator icable);
};
  

/* -------------------------------------------------------------------- */
#include "ptset/skip_range.hh"
#include "ptset/delimit.hh"
#include <strstream.h>

template <class AnyIChamber>
inline void FED<AnyIChamber>::dump(ostream& os) {
  Delimiter d("  ", 8, "\n");
  for (unsigned int icable=0; icable<size(); ++icable) {
    if ( (*this)[icable] == dummy ) os << "dummy";
    else os <<(*( (*this)[icable] )).first;
    os << d;
  }
}

template <class AnyIChamber>
inline void FEDs<AnyIChamber>::dump(ostream& os) {
  for (iterator ifed=begin(); ifed!=end(); ++ifed) {
    os << "**** FED #" << (*ifed).first << ":\n";
    (*ifed).second.dump(os);
  }
}

template <class AnyIChamber>
inline void FEDs_Filler<AnyIChamber>::operator()(istream* is) {
  unsigned int i_FED;
  if ( !(*is >> i_FED) ) {
    global_error()("can not read FED number");
    return;
  }
  FED<AnyIChamber>& fed = feds[i_FED];
  fed.dummy = cables.end();

  fill_cables(is, fed);
  if ( global_error() ) {
    string tmp; tmp << "In defining FED #" << i_FED;
    global_error() << tmp;
  }
}
    
template <class AnyIChamber>
inline void FEDs_Filler<AnyIChamber>::fill_cables(istream* is,
						  FED<AnyIChamber>& fed) {
  // trying to read N_SLOTs (cable names + Skip's)
  char c;
  string rest; // will contain the remaining part of input
  while ( is->get(c) ) rest += c;
  //  cout << "*** rest: " << rest << endl;
  
  string cable_name;
  SkipRange sr;
  Cables<AnyIChamber>::iterator icable;
  for (unsigned int islot=0; islot<FED<AnyIChamber>::N_SLOTS; ) {
    //    cout << "rest: " << rest << endl;
    strstream attempt1, attempt2;
    attempt1 << rest; attempt2 << rest;
    // attempt2 is read into cable_name, attempt1 -> into SkipRange
    rest = "";
    
    if ( attempt1 >> sr ) { // SkipRange has been read
      for (unsigned int i=0; i<sr.n; ++i) fed.push_back(cables.end());
      // insert sr.n dummy cables
      while ( attempt1.get(c) ) rest += c; // store the rest of line
      islot += sr.n;
    }
    else if ( attempt2 >> cable_name ) { // cable name has been read
      if ( (icable = cables.find(cable_name)) == cables.end() ) {
	string tmp;
	tmp << "there is no cable named \"" << cable_name
	    << "\" (FED slot #" << islot << ")";
	global_error()(tmp);
	return;
      }
      if ( (*icable).second.size()
	   != FED<AnyIChamber>::N_CHANNELS_IN_SLOT ) {
	// cable can not be inserted into FED's slot
	string tmp;
	tmp << "# channels in cable \"" << (*icable).first << "\" ("
	    << (*icable).second.size() 
	    << ")  !=  # channels in FED slot ("
	    << FED<AnyIChamber>::N_CHANNELS_IN_SLOT << ")";
	global_error()(tmp);
	return;
      }
      fed.push_back(icable);
      set_readout_flag_for_cells(icable);
      while ( attempt2.get(c) ) rest += c;
      ++islot;
    }
    else { // error
      string tmp;
      tmp << "can not read cable name #" << islot;
      global_error()(tmp);
      return;
    }
  }
  istrstream attempt(rest.c_str(), rest.size());
  attempt >> ws;
  if ( !(attempt.eof()) ) {
    string tmp;
    tmp << "more than " << FED<AnyIChamber>::N_SLOTS 
	<< " cables are given for one FED";
    global_error()(tmp);
  }
}

template <class AnyIChamber>
inline void FEDs_Filler<AnyIChamber>::set_readout_flag_for_cells
(Cables<AnyIChamber>::iterator icable) {
  Cable<AnyIChamber>& cable = (*icable).second;
  for (unsigned int i=0; i<cable.size(); ++i)
    if (cable.is_connected(i)) (*cable[i].icell).second.is_read_out = true;
}

#endif
