//-----------------------------------------------------------------------//
//
//  RingSearchARTE: ARTE interface for stand-alone ring reconstruction
//
//  Author: Reinhard Eckmann/ Denis Dujmic
//
//  Created: May-98
//
//-----------------------------------------------------------------------//
#include "rich/rich.h"
#include "rise/RingSearchARTE.h"
#include "rise/RingSearchTiming.h"
#include "arte/RSEG_cmp.hh"



#ifdef RISE_ROOT
RingSearchTiming *RICH_SearchTiming=0; 
#endif

/////////////////////////// Ring search ////////////////////////////////////////

RingSearchARTE::RingSearchARTE() : RingSearchDB() {

#ifdef RISE_TIMING
  if (!RICH_SearchTiming) RICH_SearchTiming=new RingSearchTiming;
#endif

  if (RichRun::magnet) printf(" $$$ RISE Inform: RingSearch with magnet ON\n");
  else printf(" $$$ RISE Inform: RingSearch with magnet OFF\n");

  n_mass=5;
}


void RingSearchARTE::makeHitMap() {

  int verb=0;

  // -- reset hit bank
  h.reset();

  // -- ocupancy cut
  int n0=RichRun::gede->hit1.index();
  int n1=RichRun::gede->hitl.index();
  if (n1-n0>OccupancyCut) return;

  float min_phi=0, max_phi=0;
  float min_lam=0, max_lam=0;
  if ( RingSearchTrack::exist ) {
    if (verb>1) cout<<"new ring input:"<<RingSearchTrack::phi<<" "<<RingSearchTrack::lambda<<" "<<RingSearchTrack::bx
		  <<" ("<<-RingSearchTrack::phi*4.5<<") "
	<<RingSearchTrack::by<<endl;
    RICH_LambdaPhi->Correct()->newRing(RingSearchTrack::phi, RingSearchTrack::lambda, 
				       getRadius(ArtePointer<EVHD>(1)->run), 
				       RingSearchTrack::bx,RingSearchTrack::by);
    min_phi = minSearchRadius()*.001-RingSearchTrack::max_phi_residual;
    max_phi = maxSearchRadius()*.001+RingSearchTrack::max_phi_residual;
    min_lam = minSearchRadius()*.001-RingSearchTrack::max_lambda_residual;
    max_lam = maxSearchRadius()*.001+RingSearchTrack::max_lambda_residual;
    if (verb) cout<<"min/max="<<min_phi<<" "<<max_phi<<" "<<min_lam<<" "<<max_lam<<endl;
  }


  // -- append photons
  for (int i=n0;i<n1;i++) {

    ArtePointer<HITB> hitb(i);
    
    // -- check if photon has been used
    if (!RingSearchAttrib::reuse &&
	hitb->rseg.begin()!=hitb->rseg.end()) continue;
    

    ArteRelationIterator<DIGB> idigb=hitb->digb.begin();
    ArtePointer<DIGB> digb(idigb);
    int id    = digb->id;
    int plane = (smFromChannelID(id)>6) ? 1 : 0 ;
    int size  = RICH_Detector->pm().numchannel();
    int imtra = 9999;

    // -- track seed search
    if (RingSearchTrack::exist) {
      //cout<<"RingSearchTrack::exist"<<endl;
      float dx=fabs(RICH_LambdaPhi->Correct()->dPhi   (hitb->q, hitb->r, plane));
      float dy=fabs(RICH_LambdaPhi->Correct()->dLambda(hitb->q, hitb->r, plane));
      if (dx>max_phi || dx<min_phi) continue;
      if (dy>max_lam || dy<min_lam) continue;
    }
    if (ArteTable<MTRA>::size()) { // MC mode
      ArteRelationIterator<MIMP> imimp=hitb->mimp;
      ArtePointer<MTRA> mtra=(*imimp)->mtra;
      imtra = mtra.index();
    }
    h.append(d.dualx(hitb->q*milliradian),d.dualy(hitb->r*milliradian), plane, size, imtra, hitb);
    if (verb) cout<<".";
  }
  if (verb) cout<<endl;
  if (ArteTable<MTRA>::size()) h.randomize(); 
  d.nextevent();
}


void RingSearchARTE::makeRadiusList(float r1, float r2, float dr) {

  if (radius_list.size()) radius_list.clear();

  if (RingSearchTrack::exist) {
    float oldrad=100;

    for (int i=0; i<5; i++) {
      float rr=R2P2rad2[i]*1e6;
      float R2P2sigma=1e6/sqrt(iR2P2sigma2[i]);
      if (rr<0) continue;
      float r=sqrt(rr);
      float dr=sqrt(rr+4*R2P2sigma)-r;
      if (r<minSearchRadius()) continue;
      if (oldrad-r>2*dr) radius_list.push_back(r);
      if (oldrad-r>3*dr) radius_list.push_back(r+dr); 
      if (oldrad-r>dr) { radius_list.push_back(r-dr); oldrad=r; }
    }

  }
  else {
    if (dr<0) for (float radius=r2;radius>r1;radius+=dr) {// mrad /////
      radius_list.push_back(radius);
    }
    else for (float radius=r1;radius<r2;radius+=dr) {// mrad /////
      radius_list.push_back(radius);
    }
  }

}


void RingSearchARTE::getrings(float r1,float r2,float dr,int t) {
#ifdef RISE_TIMING
  RICH_SearchTiming->startTiming();
#endif
  SetSeed();
  makeHitMap();
  makeRadiusList(r1,r2,dr);

  for (vector<float>::iterator irad=radius_list.begin();
       irad!=radius_list.end(); ++irad) {
    d.define((*irad)*1e-3*milliradian,t);
    d.ringsearch(h);
  }
  
  Choose();
  if (RingSearchAttrib::bookrseg) FillRSEG();

#ifdef RISE_TIMING
  RICH_SearchTiming->stopTiming();
  RICH_SearchTiming->fillTimingHistograms();
#endif
}


void RingSearchARTE::getrings() {
  if (!ArteTable<MTRA>::nrow()) getSearchParameters(ArtePointer<EVHD>(1)->run);
  getrings(minSearchRadius(),maxSearchRadius(),searchStep(),threshold());
}


//////////////////////////  Seed setup //////////////////////////////////

void RingSearchARTE::SetSeed() {
  //cout<<"SetSeed"<<endl;

  if (!RingSearchTrack::exist) return;

  // -- set track parameters
  RingSearchTrack::ip    = RingSearchTrack::rtra->pf;
  ArtePointer<RPNT> rpnt = (*RingSearchTrack::rtra->rpnt.begin());

  // -- get track parameters inside RICH
  //!!!!!!!!!!!!!!!!! make sure RPNT is right one !!!!!!!!!!!!!!
  float x=rpnt->xf;
  float y=rpnt->yf;
  float z=rpnt->zf;
  float tx = rpnt->txf;
  float ty = rpnt->tyf; 

  RingSearchTrack::phi    = RICH_LambdaPhi->phi   (tx, ty);
  RingSearchTrack::lambda = RICH_LambdaPhi->lambda(tx, ty);
  RingSearchTrack::bx     = ( x - tx*z )*.01; // m
  RingSearchTrack::by     = ( y - ty*z )*.01; // m

  // -- calculate expected positions and errors on R2P2 lines
  makeR2P2lines();
}

void RingSearchARTE::Choose() {
  //cout<<"RingSearchARTE::Choose()"<<endl;

  if (!RingSearchTrack::exist) return;
  // -- in case of seeded search find best ring
  float best_S=0;
  int   besti=-1;
  for (int i=0;i<numrings();i++) {
    Ring &ring=getRing(i);
    ring.accept=false;
    if (!ring.IsCloseToSeed()) continue;
    // -- find best significance or S/Sqrt(N)
    if (RingSearchAttrib::type==RingSearchType::lkbg) {
      if (best_S<ring.refit_S) {
	best_S=ring.refit_S;
	besti=i;
      }
    }
    else {
      float sn=(ring.numhit_bgrd>0) ? ring.numhit_refit/sqrt(ring.numhit_bgrd) : 100;
      if (best_S<sn) {
	best_S=sn;
	besti=i;
      }
    }
  }
  if (besti<0) return;
  Ring &bestring=getRing(besti);
  bestring.accept=true;
}



extern "C"  void rkmain_(const int    & icovar, 
              const double & zout, 
              const double & fstep, 
              const double * parameters, 
              const double * prec, 
              const double * p_track, 
              const double * derivatives, 
              const int    & ierror);

int RingSearchARTE::swim(const double *source,double *dest,const double zout) {
  const int    icovar=2;
  double       fstep=50;
  double       prec[4];
  double       derivatives[5][5];
  int          ierror;

  prec[0] = 0.001/fabs(zout-source[0]);
  prec[1] = prec[0];
  prec[2] = 0.0001/fabs(zout-source[0]);
  prec[3] = prec[2];

  rkmain_(icovar,zout,fstep,source,prec,dest,&derivatives[0][0],ierror);

  return ierror;
}


////////////////////////// booking of ARTE tables ////////////////////////


void RingSearchARTE::FillRSEG() {

  RSEG r;
  ArtePointer<RSEG>rseg;
  //cout<<"RingSearchARTE::FillRSEG() numrings="<<numrings()<<endl;
  for (int i=0; i<numrings(); i++) {

    Ring &ring=getRing(i);

    // -- check distance seed-ring center
    if (RingSearchTrack::exist && (!ring.accept) ) continue;

    // -- ring info
    r.xf = ring.phi();
    r.yf = ring.lambda();

    r.time   = ring.thetaC();
    float r0 = getRadius(ArtePointer<EVHD>(1)->run);
    r.dtim   = cos(r0)/cos(r.time);

    r.txf = ring.tx();
    r.tyf = ring.ty();

    r.pf  = float(ring.numHits());
    r.ze  = ring.numhit_refit;
    r.ye  = ring.numhit_bgrd;
    r.xe  = ring.numhit_3sigma;
    r.zf  = ring.f;
    r.txe = ring.Rb;
    r.tye = ring.refit_S;
    r.pe  = ring.sigma_signal;

    r.chi = ring.chi2;

    r.cvf[ 0] = 0;
    r.cvf[ 1] = 0;
    r.cvf[ 2] = 0;
    r.cvf[ 5] = ring.covariance(0,0);
    r.cvf[ 8] = ring.covariance(0,1);
    r.cvf[12] = ring.covariance(0,2);
    r.cvf[ 9] = ring.covariance(1,1);
    r.cvf[13] = ring.covariance(1,2);
    r.cvf[14] = ring.covariance(2,2);
    r.cvf[ 3] = 0;
    r.cvf[ 6] = 0;
    r.cvf[10] = 0;
    r.cvf[ 4] = 0;
    r.cvf[ 7] = 0;
    r.cvf[11] = 0;

    // -- component
    r.cmp = Rsegc::rich;

    // -- fit type
    if (RingSearchTrack::exist) r.fit=12;       // seeded track
    else {
      float ratio=ring.thetaC()/r0;
      ratio*=ratio;
      if (ring.numhit_refit>50*ratio) r.fit=10; // possible gamma conversion
      else r.fit=8;                             // charged track found standalone
    }

    // -- book relations
    // -- RSEG
    rseg=ArteTable<RSEG>::push_back(r);

    // -- RSEG-MTRA
    unsigned int iMtra=ring.mctrack();
    if (ArteTable<MTRA>::size() && iMtra<=ArteTable<MTRA>::size() && iMtra>0)
      createRelation(rseg,ArtePointer<MTRA>(iMtra),&RSEG::mtra);

    // -- seed based search: book RTRA-RSEG
    if (RingSearchTrack::exist) {
      createRelation(rseg, RingSearchTrack::rtra, &RSEG::rtra);
    }

    // -- book RSEG-HITB
    if (RingSearchAttrib::bookhitb) {
      for (int i=0; i<ring.numhit; i++) {
	Hit &h=(*ring.hit[i]);
	if (!h.assigned) continue;
	ArtePointer<HITB> hitb=h.photon_hitb;
	createRelation(rseg, hitb, &RSEG::hitb);
      }
    }

  }
}


void RingSearchARTE::FillRTRA(ArtePointer<RTRA> t, ArtePointer<RSEG> rich) {

  // -- fill likelihoods
  for (int i=0; i<n_mass; i++) {
    switch (i) {
    case 0: t->lre =lr[0];
    case 1: t->lrmu=lr[1];
    case 2: t->lrpi=lr[2];
    case 3: t->lrk =lr[3];
    case 4: t->lrp =lr[4];
    }
  }

  // -- fill number of rich hits
  int hri=0;
  for (ArteRelationIterator<HITB> ihitb=rich->hitb.begin();
       ihitb!=rich->hitb.end(); ++ihitb) hri++;
  t->hri=hri;

  //cout<<t<<":  lr="<<t->lre<<" "<<t->lrmu<<" "<<t->lrpi<<" "<<t->lrk<<" "<<t->lrp<<"   hri="<<t->hri<<endl;

}


/////////////////////// likelihood functions ///////////////////////////

void RingSearchARTE::makeR2P2lines() {
  float  ipip = RingSearchTrack::ip*RingSearchTrack::ip;
  float  pp   = 1/ipip;
  static float mm[]  = {2.6e-7, 1.12e-2, 1.95e-2, .24372, .88035};
  float  sigma2_ipip = 4 * ipip*ipip * (RingSearchTrack::aa+RingSearchTrack::bb*pp);

  float r0  = getRadius(ArtePointer<EVHD>(1)->run);
  float r0r0= r0*r0;
  for (int i=0; i<n_mass; i++) {
    iR2P2sigma2[i] = 1/(RingSearchTrack::sigma2_rr + mm[i]*mm[i] * sigma2_ipip);
    R2P2rad2[i]    = r0r0-mm[i]*ipip; 
  }
}



extern "C" float erfc_(float &x);

int RingSearchARTE::R2P2likelihood(ArtePointer<RTRA> t) {

  // -- find rich segment
  ArtePointer<RSEG> rich;
  ArteRelationIterator<RSEG> irseg=t->rseg.begin();
  for (;irseg!=t->rseg.end(); ++irseg) {
    if ((*irseg)->cmp==Rsegc::rich) { rich=(*irseg); break; }
  }
  if (irseg==t->rseg.end()) return -1;

  // -- fill number of rich hits
  int hri=0;
  if (RingSearchAttrib::fillrtra) { 
    for (ArteRelationIterator<HITB> ihitb=rich->hitb.begin();
	 ihitb!=rich->hitb.end(); ++ihitb) hri++;
    t->hri=hri;
  }
  
  // -- calculate likelihoods
  int ret=R2P2likelihood(rich->time);

  // -- fill tables
  if (RingSearchAttrib::fillrtra) FillRTRA(t,rich);

  return ret;
}



int RingSearchARTE::R2P2likelihood(float radius) {

  int verb=0;

  // -- calculate deviations from r2p2 lines
  float chi2=20;
  memset(&lr[0],0,sizeof(float)*n_mass);
  float rr=radius*radius;
  int bestM=-1;
  for (int i=0; i<n_mass; i++) {
    float d       = (rr - R2P2rad2[i]);
    lr[i]         = d*d*iR2P2sigma2[i];
    if (lr[i]<chi2) { chi2=lr[i]; bestM=i; }
  }

  float lrsum=0;
  if (bestM>-1) {
    // -- calculate probabilities
    for (int i=0; i<n_mass; i++) {
      if (verb) cout<<i<<":  "<<lr[i]<<" ";
      if (lr[i]>12) lr[i]=0;
      else {
	lr[i]=sqrt(lr[i]);
	lr[i]=erfc_(lr[i]);
	lrsum+=lr[i];
      }
      if (verb) cout<<lr[i]<<" "<<endl;
    }
    if (verb) cout<<endl;

    // -- renormalize
    if (lrsum>1e-6) for (int i=0; i<n_mass; i++) lr[i]/=lrsum;

  }
  else memset(&lr[0],0,sizeof(float)*n_mass);

  return bestM;
}



RingSearchARTE *RICH_Search=0;

//////////////////////// functions called by user ////////////////////////////

extern "C" {

void rcrings_(float r1,float r2,float dr,int t) {
  RingSearchTrack::exist     = false;
  RingSearchAttrib::reuse    = true;
  RingSearchAttrib::bookrseg = true;
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RICH_Search->getrings(r1,r2,dr,t);
}

void rcringsdb_() {
  RingSearchTrack::exist     = false;
  RingSearchAttrib::reuse    = true;
  RingSearchAttrib::bookrseg = true;
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RICH_Search->getrings();
}

void rcringsid_(ArtePointer<RTRA> track) {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RingSearchTrack::exist     = true;
  RingSearchTrack::rtra      = track;
  RingSearchAttrib::reuse    = true;
  RingSearchAttrib::bookrseg = true;
  RingSearchAttrib::bookhitb = true;
  RICH_Search->getrings();
  RICH_Search->R2P2likelihood(track);
}

void rcringsclean_() {
  RingSearchTrack::exist     = false;
  RingSearchAttrib::reuse    = false;
  RingSearchAttrib::bookrseg = true;
  RingSearchAttrib::bookhitb = false;
  if (!RICH_Search) {
    fprintf(stderr," $$$ RICH cleanup: Ring cleanup called w/o previous call to ringseach\n");
    return;
  }
  RICH_Search->getrings();
}
	   }


float* ringid(ArtePointer<RTRA> track) {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RingSearchTrack::ip        = track->pf;
  RingSearchAttrib::fillrtra = true;
  RICH_Search->makeR2P2lines();
  RICH_Search->R2P2likelihood(track);
  return RICH_Search->GetLr();
}

float* ringid(float pf, float radius) {
  if (!RICH_Search) RICH_Search=new RingSearchARTE;
  RingSearchTrack::ip        = pf;
  RingSearchAttrib::fillrtra = false;
  RICH_Search->makeR2P2lines();
  RICH_Search->R2P2likelihood(radius);
  return RICH_Search->GetLr();
}
