// ********************************************************************
//
//  Routine:      acReadVDS
//  Author:       T.Glebe, I.Kisel
//
//  Description:  call the VDS initialization to get the VDS constants
//                - if key not assigned, use run number as major number
//                - returns <0 error
//                           0 no VDS update necessary
//                           1 update done using key
//                           2 update done using date/time or run number
//                
//
//  Modification:
//
// ********************************************************************
//!!!!!!!!!!!!!!!!!!!!!!
using namespace std;//!!
//!!!!!!!!!!!!!!!!!!!!!!

#include <stdio.h>

#include "arte/arte.hh"
#include "database/keytable.h"
#include "farm/farm.hh"
#include "grr.h"
#include "silarte/silarte.hh"
#include "vdsleda.h"

// sparse event definition
#include "event.h"
#include "etb.h"

// prototypes
int SvdStartup();
extern "C" {
  void fill_cats(void);
}

static int VDSserver = -1;
static int VDStable  = -1;


//*****************************************************************
// function
//*****************************************************************

int acReadVDS(int KTstatus) 
{
  
  //==================================================================
  // no action for MC
  //==================================================================
  
  if (ArteTable<MTRA>::size() > 0 ) return 0;
  
//==================================================================
// database server (default: /VDS_CONFIG)
//                  can be changed by setting DB_VDS
//==================================================================
 
  char defaultserver[13];
  sprintf(defaultserver,"/VDS_CONFIG");
  char * dbserver;
  
  if (getenv("DB_VDS") == NULL){
    dbserver = defaultserver;
  } else {
    dbserver = getenv("DB_VDS");
  };
 
  
  switch (KTstatus) { 
    case 2: 
      // structure of keytable changed, so loop over the keytable and get 
      // your serverid and tableids 

      VDSserver = KeyFindServer(dbserver); 
      if (VDSserver == -1) { 
        cout<<"acReadVDS: KeyTable Server not found!"<<endl; 
        //return -1; 
	break;  // use old method
      }; 
      //      VDStable = KeyFindTable(VDSserver, "VDS/Side"); 
      VDStable = KeyFindTable(VDSserver, "Side"); 
      if (VDStable == -1) { 
        cout<<"acReadVDS: Keytable not found!"<<endl; 
        //return -1; 
	break;  // use old method
      }; 
      // NO _BREAK_ HERE! 

    case 1: 
      // an update happened, look if your server is involved 
      if (!KeyServerNew(VDSserver)) 
	return 0; //no VDS update necessary
      // now for the tables - look whether they are changed 
      if (KeyTableNew(VDStable)) { 
      // using LEDA to get the Tables x:
        MizziDatabase(mydb);
        Tables x;
        mydb.open(KeyServerName(VDSserver),DBreadonly);
        x=mydb.get_id();
        VDS_Side->book(mydb);
        KeyTableStamp(x,KeyServerName(VDSserver),KeyTableName(VDStable));
        VDS_Side   ->fetch_data( DBle );
        VDS_Side   ->set_keys();
        VDS_Side   ->match( DBle );
        VDS_Side   ->unbook();

        silarte_updateGEDE();
        fill_cats(); //fill cats with new geometry
	return 1; // update using keytable
      } 
      return 0;  // no VDS update necessary
      //break; 
  } // of switch 

  // No keytable information available, using old method...

  static bool firsttime = true;
  if(firsttime) {
    cout << "acReadVDS: Key NOT assigned --> use old method" << endl; 
    SvdStartup();
    firsttime = false;
  }

  static int VDS_Side_major = -1;
  static int VDS_Side_minor = -1;
  static bool findmajmin = true;
  static MizziDatabase( vdsdb_setup );

  // get event pointer
  EVENT_HDR *event = ARTE_get_event();
  if ( !event ) {
    cout << "acReadVDS: Could not find event!" << endl;
    return -1; // error
  }      

  //    Get constants from database
  //
  // Time stamp is evaluated in this order:
  //  1. date/time   from vdsleda IOL
  //  2. major/minor from vdsleda IOL
  //  3. event time from DDAQ (if all IOL parameters are =0)
  int major = 0;
  int minor = 0;
  int ndaq;

  EVT_DATA_RECORD daq;
  ETB_get_daqs( event, 1, &daq, &ndaq );
  if ( ndaq > 0 ) {
    major = daq.data.event_time;
    minor = 0;
  }

  if(VDS_Side_major == -1 || 
     (major > VDS_Side_major && !findmajmin)) {

    vdsleda_fetchdata(major, minor);
    silarte_updateGEDE();
    fill_cats(); //fill cats with new geometry
  } else {
    return 0; // no update necessary
  }

  // find major minor until which dataset is valid
  if(VDS_Side_major == -1  || 
     findmajmin            ||
     major > VDS_Side_major ) {

    //    vdsdb_setup.open( vdsleda_iol.setup_database.value, 0 );
    vdsdb_setup.open( dbserver, 0 );
    vdsdb_setup.stamp( major, minor );
    VDS_Side->book( vdsdb_setup );
    if( VDS_Side->match( DBgt ) == 1 ) {
      // successfull match
      VDS_Side_major = TableMajor( vdsdb_setup.get_id() );
      VDS_Side_minor = TableMinor( vdsdb_setup.get_id() );
      findmajmin = false;
    } else {
      //if there is no newer table, lookup would take place in each event!
      //      findmajmin = true;
      VDS_Side_major = major + 200;
      findmajmin = false;
    }

    VDS_Side->unbook();
    vdsdb_setup.close();
  }

  return 2;  // constants read in using old method 
}
