// *****************************************************************************
//
// source:    L3_Geometry.C
//
// type:      source code
//
// created:   06 Dec 1999
//
// author:    Stefan Scharein, Ullrich Schwanke
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: scharein@ifh.de, schwanke@ifh.de
//
// symbol:    
//
// changes:
//
// *****************************************************************************
#include "L3_Geometry.hh"
#include "L3_Parameter.hh"
#include "L3_VTstruct.hh"
#include "L3_Vtx.hh"
#include <cmath>
#include <cstdio>
#include <algorithm>

#ifdef L3_SLP_FRAME
#include <cstdlib>
extern "C" {
#include "l2geo.h"
}
#endif

vector<L3_Target> L3_Geometry::targetList; //list with targets (i.e. target wires)
vector<L3_Target*> L3_Geometry::targetListZ;  
float L3_Geometry::the_maxztarget;
float L3_Geometry::the_minztarget;
float L3_Geometry::the_maxxtarget;
float L3_Geometry::the_minxtarget;
float L3_Geometry::the_maxytarget;
float L3_Geometry::the_minytarget;
float L3_Geometry::the_TargetBox_xmin;
float L3_Geometry::the_TargetBox_xmax;
float L3_Geometry::the_TargetBox_ymin;
float L3_Geometry::the_TargetBox_ymax;
float L3_Geometry::the_TargetBox_zmin;
float L3_Geometry::the_TargetBox_zmax;
L3_Plane L3_Geometry::SiVD_PlaneList[7][4][4];
float L3_Geometry::SuperLayer_z[7][4];
L3_Window L3_Geometry::secs[7][4]; 
Hep3Vector L3_Geometry::the_mean_target;
int L3_Geometry::LUT[7][4][4][3];
vector<L3_TargetStation> L3_Geometry::stationList;
int L3_Geometry::numPlanes[7][4];
int L3_Geometry::sectorAtSameZPosition[7][4];

void L3_Geometry::initialize() {
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Geometry::initialize() has been called" << endl;
#endif
  fillTargetGeometry();
  fillVDSGeometry();
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Geometry::initialize() is done" << endl;
#endif
}

void L3_Geometry::updateTarget(){
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()){
    for(vector<L3_Target>::iterator i=targetList.begin();i!=targetList.end();i++){
      printf("Target %i (%s) %f, %f, %f\n",i-targetList.begin()+1,i->name(),i->x(),i->y(),i->z());
    }
  }
#endif
  targetList.clear();
  targetListZ.clear();
  fillTargetGeometry();
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()){
    for(vector<L3_Target>::iterator i=targetList.begin();i!=targetList.end();i++){
      printf("Target %i (%s) %f, %f, %f\n",i-targetList.begin()+1,i->name(),i->x(),i->y(),i->z());
    }
  }
#endif
}

void L3_Geometry::update() {
//in SLP it might happen that we are called again because new target positions
//were read into the SLP, so we clear all existing target wires
  targetList.clear();
  targetListZ.clear();
  stationList.clear();
  fillTargetGeometry();
  fillVDSGeometry();
}

void L3_Geometry::fillTargetGeometry() {
  readTargets();
  calculate_TargetPosition();
  calculate_TargetBox();
}

void L3_Geometry::fillVDSGeometry() {
  readPlanes();
  calculate_mean_z();
  calculate_sectors();
}

void L3_Geometry::readTargets() {
#ifdef L3_VERBOSE
  if( L3_Parameter::Trace() ) cout << "L3_Geometry::readTargets() START" << endl;
#endif
#ifdef L3_ARTE_FRAME
  // in older ARTE versions it might be that all GTAR.eff <= 0.0 -> then we ignore it
  int flag = 0; 
  for(ArteTable<GTAR>::iterator t=ArteTable<GTAR>::begin();
      t!=ArteTable<GTAR>::end();t++) {
    if(t->eff > 0.0){ flag=1; break; }
  }
  for(ArteTable<GTAR>::iterator t=ArteTable<GTAR>::begin();
      t!=ArteTable<GTAR>::end();t++) {
    L3_Target wire(*t);
    int take = 1;
    if(!ArteTable<MCIN>::booked()){
      //for real data we check whether the wire is on the beam - 
      //according to D. Samtleben
      if( pow(wire.x(),2)+pow(wire.y(),2) >= 1.0 ) take = 0;
    } else if(flag) { 
      //for MC data we skip wires with eff<=0.0
      if( t->eff <= 0 ) take = 0;
    }
    if( take ){
      targetList.push_back( wire );
    } else {
#ifdef L3_VERBOSE
      if(L3_Parameter::Info()) cout << "Wire " << wire.name() << " is not on the beam...." << endl;
#endif
    }
  }
#endif
#ifdef L3_SLP_FRAME
  for(const L2TARG *wire = l2targ_IterateInit(); wire!=NULL; wire = l2targ_IterateNext() ) {
    targetList.push_back( L3_Target(*wire) );
  }
#endif
#ifdef L3_VERBOSE
  if( L3_Parameter::Info() ){
    cout << "L3_Geometry::readTargets: Read " << nTarget() << " target wires" << endl;
    for(vector<L3_Target>::const_iterator i = targetList.begin();i!=targetList.end();i++) {
      cout << *i;
    }
  }
#endif
  { //------------ fill target information into VT's common block -----------
    int i;
    for(i=0;i<nTarget();i++)
      {
	const L3_Target &targ = getTarget(i);
	int verti = 0; //assume horizontal wire
	const double pihalf=M_PI_2;
	const double sqrt12=sqrt(12.0);
	// Wire orientation
	if( targ.dy() > targ.dx() ) {
	  if( targ.x() < 0 ) 
	    verti = -1;
	  else
	    verti = 1;
	} 
	// Wire Position and rotation angles
	vtwire_.xw[i]   = Point( targ.x(),targ.y(),targ.z() );
	vtwire_.thew[i] = 0.0;
	vtwire_.phiw[i] = verti*pihalf;
	// RMS Width and length of wire (in beam direction)
	double sigmaXY = L3_Parameter::sigmaxy();
	if( sigmaXY <= 0.0 ) sigmaXY = (targ.dx()<targ.dy()) ? targ.dx()/sqrt12 : targ.dy()/sqrt12;
	vtwire_.sigtw[i] = sigmaXY;
	double sigmaZ = L3_Parameter::sigmaz();
	if( sigmaZ<=0.0 ) sigmaZ = targ.dz()/sqrt12; 
	vtwire_.siglw[i] = sigmaZ;
	if( L3_Parameter::Info() ){
	  int sigxy=int(1E4*sigmaXY+0.5);
	  int sigz =int(1E4*sigmaZ+0.5);
	  cout<<"L3_Geometry::init: "<<targ.name()<<" sigmaxy/z = "<<sigxy<<'/'<<sigz<<" microns"<<endl;
	}
      }
    vtwire_.nw = i;
    //cout << vtwire_.nw << " wires in buffer" << endl;
    i=0; //initialize all wires
    vtwini_(i);
    //cout << "done" << endl;
  }
#ifdef L3_VERBOSE
  if( L3_Parameter::Trace() ) cout << "L3_Geometry::readTargets() STOP" << endl;
#endif
}

void L3_Geometry::showGeometry()
{
  char buf[] = "!!";
  for(int i=0;i<9*7+7;i++) cout << "-"; cout << endl;
  cout << "Sector ";
  for(int sl=0;sl<7;sl++) printf("  SL%02i   ",sl+1);     
  cout << endl;
  for(int i=0;i<9*7+7;i++) cout << "-"; cout << endl;
  for(int sec=0;sec<4;sec++)
    {
      cout << "   " << sec+1 << "   ";
      for(int sl=0;sl<7;sl++){
	for(int pl=0;pl<4;pl++){
	  const L3_Plane& p = L3_Geometry::getPlane(sl,sec,pl);
	  if( p.l3nsl() == -1 ){
	    cout << "  ";
	  } else {
	    buf[0] = (atan(p.sin()/p.cos())>0) ? '+' : '-';
	    buf[1] = (fabs(p.cos())>0.5) ? 'y' : 'x';
	    cout << buf;
	  }
	}
	cout << ' ';
      }
      cout << endl;
    }
  for(int i=0;i<9*7+7;i++) cout << "-"; cout << endl;  
}

int L3_Geometry::readPlanes() {
  int foundSomething = 0;
  for (int SL=0; SL<7; SL++) {
    for (int Sec=0; Sec<4; Sec++) {
      for (int Pl=0; Pl<4; Pl++) {
	L3_Plane l3plane( 0.5, -1, (SL*16+Sec*4+Pl+1) );
	SiVD_PlaneList[SL][Sec][Pl] = l3plane;
      }
    }
  }
#ifdef L3_ARTE_FRAME
  if( ArteTable<GEHD>::booked() && L3_Parameter::Debug() ){
    ArtePointer<GEHD> gehd(1);
    cout << "L3_Geometry: version is " << gehd->geom << endl;
  } 
  for (ArteTable<GEDE>::iterator PIt=ArteTable<GEDE>::begin(); 
       PIt!=ArteTable<GEDE>::end(); ++PIt ) {
    ArtePointer<GEDE> Plane(*PIt);
    // Unpack SiVD GEDE-Planes, if packed
    if(Plane->sec > 65000) {
      Plane->sec = Plane->sec & 0xFFFF;
    }
    if(Plane->lay > 65000) {
      Plane->lay = Plane->lay & 0xFFFF;
    }  
    ArtePointer<GESL> agesl(Plane->gesl);
    int SL = agesl.index();
    if ( (Plane->cmp == 1) && (Plane->zfw[0] < 180) && (SL > 0 )) {
      SL = agesl->nsl;
      int l3gede = Plane.index();
      L3_Plane l3plane(Plane, SL, l3gede);
      if( SiVD_PlaneList[l3plane.l3nsl() -1][l3plane.sec() -1][l3plane.lay() -1].l3nsl() != -1 )
	{
	  cout << "L3_Geometry::readPlanes: WARNING (sl,sec,lay)=(";
	  cout << l3plane.l3nsl() << ',' << l3plane.sec() << ',' << l3plane.lay() << ") is overwritten!!" << endl;
	}
      SiVD_PlaneList[l3plane.l3nsl() -1][l3plane.sec() -1][l3plane.lay() -1] = l3plane;
      //cout << l3plane.l3nsl() << ' ' << l3plane.sec() << ' ' << l3plane.lay() << ' ' << l3plane.cos() << ' ' << l3plane.sin();
      //cout << ' ' << atan( l3plane.sin()/l3plane.cos() )*180.0/M_PI << ' ';
      //cout << l3plane.sxmn() << ".." << l3plane.sxmx() << "  " << l3plane.symn() << ".." << l3plane.symx() << endl;
      //cout << l3plane.l3nsl() << ' ' << l3plane.sec() << ' ' << l3plane.lay() << ' ' << Plane->eff << endl;
      foundSomething = 1;
    }
  }
#endif
#ifdef L3_SLP_FRAME
  char nameSL[6];
  for(int iSL = 1; iSL < 8; iSL++) { // for each superlayer in the vertex detector
    sprintf(nameSL,"SI%02i",iSL);
    for (int Sector = 0; Sector < 4; Sector++) { // for each Sector
      int nLayers = 0;
      // get pointers to Layers
      PLANE** plArray = Geo_GetPlanes(nameSL, Sector+1, &nLayers);      
      cout << "L3: " << nameSL << ' ' << Sector+1 << ' ' << nLayers << ' ' << (void*)plArray << endl;
      if ( plArray != NULL) {
	if (nLayers != 0) {
	  foundSomething = 1;
	  for (int Layer = 0;Layer<nLayers;Layer++) { // for each Layer
	    if( Layer>=4 ){
	      cout << "Error when reading geometry, layer = " << Layer << endl;
	      continue;
	    }
            PLANE *siPl = plArray[Layer];
	    int l3gede = 16*(iSL-1)+ 4*Sector + Layer + 1;
	    L3_Plane l3plane(siPl, iSL, l3gede);
	    if( SiVD_PlaneList[l3plane.l3nsl() -1][l3plane.sec() -1][l3plane.lay() -1].l3nsl() != -1 )
	      {
		cout << "L3_Geometry::readPlanes: WARNING (sl,sec,lay)=(";
		cout << l3plane.l3nsl() << ',' << l3plane.sec() << ',' << l3plane.lay() << ") is overwritten!!" << endl;
	      }
	    SiVD_PlaneList[l3plane.l3nsl() -1][l3plane.sec() -1][l3plane.lay() -1] = l3plane;
	  }	    
	}
        free(plArray);
      }
    }
  }
#endif  
#ifdef L3_VERBOSE
  if(L3_Parameter::Info()) showGeometry();
#endif
  calculateLUT();
  return foundSomething;
}

void L3_Geometry::calculate_TargetPosition() {
  //calculate target position for constraints  
  if( nTarget() > 0 ) {
    the_minztarget = targetList[0].z();
    the_maxztarget = the_minztarget;
    the_minytarget = targetList[0].y();
    the_maxytarget = the_minytarget;
    the_minxtarget = targetList[0].x();
    the_maxxtarget = the_minxtarget;
    the_mean_target = targetList[0].get3center();
    for(vector<L3_Target>::const_iterator i = targetList.begin()+1;i!=targetList.end();i++)
      {
	if( i->z() < the_minztarget ) the_minztarget = i->z();
	if( i->z() > the_maxztarget ) the_maxztarget = i->z();
	if( i->y() < the_minytarget ) the_minytarget = i->y();
	if( i->y() > the_maxytarget ) the_maxytarget = i->y();
	if( i->x() < the_minxtarget ) the_minxtarget = i->x();
	if( i->x() > the_maxxtarget ) the_maxxtarget = i->x();
	the_mean_target += i->get3center();
      }
    the_mean_target *= 1.0/float(nTarget());
  }
  for(int i=0;i<nTarget();i++)
    {
      targetListZ.push_back( &(targetList[i]) );
    }
  //sort target list in z
  for(int i=0;i<nTarget();i++)
    {
      int min = i;
      for(int k=i+1;k<nTarget();k++) if( targetListZ[k]->z() < targetListZ[min]->z() ) min=k;
      if( min!=i )
	{
	  L3_Target *tmp   = targetListZ[i];
	  targetListZ[i]   = targetListZ[min]; 
	  targetListZ[min] = tmp;
	}
    }  
  //fill the list of target stations
  {
    struct { int station; float z; } xx[2] = {{0,0.0},{0,0.0}};
    for(vector<L3_Target>::iterator i=targetList.begin();i!=targetList.end();i++){
      //cout << i->wnr() << ' ';
      //1 2 3 4 -> station II -> index = 1
      //5 6 7 8 -> station I  -> index = 0
      int index = (i->wnr()<=4) ? 1 : 0;
      xx[index].station++;
      xx[index].z += i->z();
    }
    //cout << endl;
    for(int i=0;i<2;i++) if( xx[i].station > 0 ){
      float meanz = xx[i].z/float(xx[i].station);
      stationList.push_back( L3_TargetStation(meanz,i+1) );
      if( L3_Parameter::Trace() ) cout << "L3_Geometry: target station " << i+1 << " at z=" << meanz << endl;
    }
  }
#ifdef L3_VERBOSE
  if(L3_Parameter::Info()){
    cout << "L3_Geometry::TargetPosition: zmintarget = " << the_minztarget << endl;
    cout << "L3_Geometry::TargetPosition: zmaxtarget = " << the_maxztarget << endl;
    cout << "L3_Geometry::TargetPosition: ymintarget = " << the_minytarget << endl;
    cout << "L3_Geometry::TargetPosition: ymaxtarget = " << the_maxytarget << endl;
    cout << "L3_Geometry::TargetPosition: xmintarget = " << the_minxtarget << endl;
    cout << "L3_Geometry::TargetPosition: xmaxtarget = " << the_maxxtarget << endl;
    cout << "L3_Geometry::TargetPosition: xtarget    = " << the_mean_target.x() << endl;
    cout << "L3_Geometry::TargetPosition: ytarget    = " << the_mean_target.y() << endl;
    cout << "L3_Geometry::TargetPosition: ztarget    = " << the_mean_target.z() << endl;
  }
#endif
}

void L3_Geometry::calculate_TargetBox() {
  the_TargetBox_xmin = 0.0 - fabs(0.5 * L3_Parameter::TargetBox_Width());
  the_TargetBox_xmax = 0.0 + fabs(0.5 * L3_Parameter::TargetBox_Width());
  the_TargetBox_ymin = 0.0 - fabs(0.5 * L3_Parameter::TargetBox_Width());
  the_TargetBox_ymax = 0.0 + fabs(0.5 * L3_Parameter::TargetBox_Width());
  the_TargetBox_zmin = the_minztarget - fabs(L3_Parameter::Offset_TargetBox_zmin());
  the_TargetBox_zmax = the_maxztarget + fabs(L3_Parameter::Offset_TargetBox_zmax());
#ifdef L3_VERBOSE
  if( L3_Parameter::Trace() ){
    cout << "L3_Geometry::TargetBox: xmin ... xmax : " << the_TargetBox_xmin;
    cout << " ... " << the_TargetBox_xmax << endl;
    cout << "L3_Geometry::TargetBox: ymin ... ymax : " << the_TargetBox_ymin;
    cout << " ... " << the_TargetBox_ymax << endl;
    cout << "L3_Geometry::TargetBox: zmin ... zmax : " << the_TargetBox_zmin;
    cout << " ... " << the_TargetBox_zmax << endl;
  }
#endif
}

void L3_Geometry::calculate_mean_z() {
  for (int sl=0;sl<7;sl++) {
    for (int sec=0;sec<4;sec++) {
      float nPlanes = 0.0;
      float z = 0.0;
      for(int pl=0;pl<4;pl++){
	const L3_Plane& Plane = getPlane(sl,sec,pl);
	if( Plane.l3nsl()!=-1 ){
	  z += Plane.zfw(0);
	  nPlanes = nPlanes + 1.0;
	}
      }
      if(nPlanes>0.0) SuperLayer_z[sl][sec] = z/nPlanes;
      //if( L3_Parameter::Info() ) cout << "SLz[" << sl+1 << "][" << sec+1 << "] = " << SuperLayer_z[sl][sec] << endl;
    }
  }
}

void L3_Geometry::calculate_mean_z_old() {
  for (int sl=0;sl<7;sl++) {
    for (int sec=0;sec<4;sec++) {
      float zmin = +10000.0;
      float zmax = -10000.0;
      for(int pl=0;pl<4;pl++){
	const L3_Plane& Plane = getPlane(sl,sec,pl);
	if(Plane.zfw(0)<zmin && Plane.l3nsl() != -1) {
	  zmin = Plane.zfw(0);
	}
	if(Plane.zfw(0)>zmax && Plane.l3nsl() != -1) {
	  zmax = Plane.zfw(0);
	}
      }      
      SuperLayer_z[sl][sec] = 0.5 * (zmin + zmax);
      //if( L3_Parameter::Info() ) cout << "SLz[" << sl+1 << "][" << sec+1 << "] = " << SuperLayer_z[sl][sec] << endl;
    }
  }
}

void L3_Geometry::calculate_sectors() {
  const float d = 0.1; //allow for 1 mm deviation
  const float maxdz = 0.05;
  for(int sl=0;sl<7;sl++)
    {
      for(int sec=0;sec<4;sec++)
	{
	  //calculate windows
	  int found = 0;
	  float xmin=0,xmax=0,ymin=0,ymax=0;
	  for(int pl=0;pl<4;pl++)
	    {
	      const L3_Plane& p = getPlane(sl,sec,pl);
	      if(p.l3nsl()!=-1){
		found++;
		if( found == 1 )
		  {
		    xmin = p.sxmn();
		    xmax = p.sxmx();
		    ymin = p.symn();
		    ymax = p.symx();
		  }
		else
		  {
		    if( p.sxmn() < xmin ) xmin = p.sxmn();
		    if( p.sxmx() > xmax ) xmax = p.sxmx();
		    if( p.symn() < ymin ) ymin = p.symn();
		    if( p.symx() > ymax ) ymax = p.symx();
		  }
	      }		
	    }
	  if(found>0)
	    secs[sl][sec] = L3_Window(xmin-d,xmax+d,ymin-d,ymax+d,SLz(sl,sec));
	  else
	    secs[sl][sec] = L3_Window(-1000,-1000,-1000,-1000,-1000);
	  //load lookup-table 
	  int sector = 0;
	  for(;sector<sec;sector++){
	    if( fabs(SLz(sl,sector)-SLz(sl,sec))<maxdz ) break;
	  }
	  sectorAtSameZPosition[sl][sec]=sector;
	}
    }
}

int L3_Geometry::NPNPorNPPN(int sl,int sec){
  for(int p1=0;p1<2;p1++){
    const L3_Plane& pl1 = getPlane(sl,sec,p1);
    if( pl1.l3nsl() == -1 ) continue;
    int sign1 = (fabs(pl1.cos())>0.5) ? -1 : 1;
    for(int p2=p1+1;p2<4;p2++){
      const L3_Plane& pl2 = getPlane(sl,sec,p2);
      if( pl2.l3nsl() == -1 ) continue;
      int sign2 = (fabs(pl2.cos())>0.5) ? -1 : 1;
      if( sign1==sign2 ){
	if( (p2-p1)==2 ) //planes with the same stereo angle are separated by 2 planes
	  return +1; //NPNP
	else
	  return -1; //NPPN    
      }
    }
  } 
  return 0; //no decision
}

void L3_Geometry::calculateLUT()
{
  char *s=NULL;  
  int Default = 0;
  for(int sl=0;Default==0 && sl<7;sl++) for(int sec=0;Default==0 && sec<4;sec++) Default = NPNPorNPPN(sl,sec);
  if( L3_Parameter::Info() ){
    switch(Default){
    case +1 : cout << "L3_Geometry: NPNP" << endl; break;
    case -1 : cout << "L3_Geometry: NPPN" << endl; break;
    default : cout << "L3_Geometry: ????" << endl; break;
    }
  }
  if( L3_Parameter::Info() ){
    cout << "SL   sector1 sector2 sector3 sector4" << endl;
    cout << "------------------------------------" << endl;
    //       SI01 NPPN    NPPN    NPPN    NPPN
  }
  for(int sl=0;sl<7;sl++) 
    {
      if( L3_Parameter::Info() ) cout << "SI0" << sl+1 << " ";
      for(int sec=0;sec<4;sec++){
	numPlanes[sl][sec]=0;
	int type = NPNPorNPPN(sl,sec);
	if(type==0) type=Default;
	for(int pl=0;pl<2;pl++){
	  if( type==1 ){  //NPNP
	    s = "NpNp";
	    LUT[sl][sec][pl][0] = 2+pl; //stereo plane
	    LUT[sl][sec][pl][1] = 1-pl; //1st other plane
	    LUT[sl][sec][pl][2] = 3-pl; //2nd other plane
	    LUT[sl][sec][pl+2][0] = pl; //stereo plane
	    LUT[sl][sec][pl+2][1] = 1-pl; //1st other plane
	    LUT[sl][sec][pl+2][2] = 3-pl; //2nd other plane
	  } else {        //NPPN
	    s = "NppN";
	    LUT[sl][sec][pl][0] = 3-pl; //stereo plane
	    LUT[sl][sec][pl][1] = 1-pl; //1st other plane
	    LUT[sl][sec][pl][2] = 2+pl; //2nd other plane
	    LUT[sl][sec][pl+2][0] = 1-pl; //stereo plane
	    LUT[sl][sec][pl+2][1] = pl; //1st other plane
	    LUT[sl][sec][pl+2][2] = 3-pl; //2nd other plane
	  }
	  if( getPlane(sl,sec,pl  ).l3nsl() != -1 ) numPlanes[sl][sec]++;
	  if( getPlane(sl,sec,pl+2).l3nsl() != -1 ) numPlanes[sl][sec]++;
	}
	if( L3_Parameter::Info() )  cout << s << '(' << numPlanes[sl][sec] << ") ";	
      }
      if( L3_Parameter::Info() ) cout << endl;
    }
  if( L3_Parameter::Info() ) cout << "------------------------------------" << endl;
}

const L3_Target& L3_Geometry::ClosestTarget( Hep3Vector v )
{
  const L3_Target *target = &L3_Geometry::getTarget(0);  
  float dd = (target->get3center()-v).mag();
  for(int i=1;i<L3_Geometry::nTarget();i++)
    {
      const L3_Target *tmp = &L3_Geometry::getTarget(i);
      float newdd = (tmp->get3center()-v).mag();
      if( newdd < dd )
	{
	  target = tmp;
	  dd = newdd;
	}
    }
  return *target;
}



