// *****************************************************************************
//
// source:    L3_Monitor.C
//
// type:      source code
//
// created:   06 Dec 1999
//
// author:    Stefan Scharein, Ullrich Schwanke
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: scharein@ifh.de, schwanke@ifh.de
//
// symbol:    
//
// changes:
//
// *****************************************************************************
#include "L3_Monitor.hh"
#ifndef L3_NO_RHP
#include "L3_State.hh"
#include "L3_Event.hh"
#include "L3_Track.hh"
#include "L3_SpacePoint.hh"
#include "L3_Hitb.hh"
#include "L3_Timer.hh"
#include <stdio.h>
#include "rhp.h"

L3_Monitor::L3_Monitor() {
}

L3_Monitor::L3_Monitor(const L3_Monitor& a) {
}

L3_Monitor& L3_Monitor::operator = (const L3_Monitor& a) {
  return *this;
}

void L3_Monitor::init() {
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Monitor::init() has been called" << endl;
#endif
  char buf[128];
  //overall information
  rhp_hbook1(MYID(6000), "TLT: Number of hits in SVD", 100, 0.0, 500.0);
  rhp_hbook1(MYID(6001), "TLT: Number of spacepoints reconstructed", 100, 0.0, 500.0);
  rhp_hbook1(MYID(6002), "TLT: Number of tracks reconstructed", 25, 0.0, 25.0);
  rhp_hbook1(MYID(6003), "TLT: Number of primary vertices reconstructed", 8, -0.5,-0.5+8);

  //spacepoints - one histogram per superlayer
  for(int sl=1;sl<8;sl++){
    sprintf(buf,"TLT: spacepoints SL%02i",sl);
    int ID = 7000 + sl;
    rhp_hbook2(MYID(ID),buf,100, -10.0, 10.0, 100, -10.0,10.0);
  }
  //tracks - one histogram showing phase space coverage
  rhp_hbook2(MYID(8000),"TLT: tracks at ECAL",100,-200,200, 100,-200,200);

  //primary vertices
  const float def = 99999;
  float minx = def,maxx=-def;
  float miny = def,maxy=-def;
  int nTarget = L3_Geometry::nTarget();
  for(int i=0;i<nTarget;i++){
    const L3_Target &t = L3_Geometry::getTarget(i);
    float minX=0,maxX=0,minY=0,maxY=0;
    if( t.dx() > t.dy() ){  //horizontal wire
      maxY = t.y()+0.05;
      minY = t.y()-0.05;
      maxX = t.x()+0.4;
      minX = t.x()-0.4;
    } else {
      maxY = t.y()+0.4;
      minY = t.y()-0.4;
      maxX = t.x()+0.05;
      minX = t.x()-0.05;
    }
    if( maxX>maxx ) maxx=maxX;
    if( maxY>maxy ) maxy=maxY;
    if( minX<minx ) minx=minX;
    if( minY<miny ) miny=minY;
  }
  if( L3_Parameter::Info() ){
    printf("L3_Monitor %f<=x<=%f\n",minx,maxx);
    printf("L3_Monitor %f<=y<=%f\n",miny,maxy);
    printf("L3_Monitor %f<=z<=%f\n",L3_Geometry::minztarget()-0.1,L3_Geometry::maxztarget()+0.1);
  }
  rhp_hbook1(MYID(9000),"TLT: xprimvertex",100,minx,maxx);
  rhp_hbook1(MYID(9001),"TLT: yprimvertex",100,miny,maxy);  
  rhp_hbook1(MYID(9002),"TLT: zprimvertex",100,L3_Geometry::minztarget()-0.1,L3_Geometry::maxztarget()+0.1);
  rhp_hbook1(MYID(9003),"TLT: vertices per wire",8,0.5,0.5+8);
  
#ifdef L3_VERBOSE
  if(L3_Parameter::Trace()) cout << "L3_Monitor::init() is done" << endl;
#endif
}

void L3_Monitor::put(const L3_Event& Event) {
  put((Event.silicon()).Tracks());
  put((Event.silicon()).Points());
  put((Event.origin()).primVertices());
  rhp_hf1(MYID(6000), Event.silicon().nHitbs() ,1);
}

void L3_Monitor::put(const L3_State& State) {
}

void L3_Monitor::put(const L3_Timer& Timer) {
}

void L3_Monitor::stop() {
}

void L3_Monitor::put(vector<L3_Track> TrackList) {
  rhp_hf1(MYID(6002), TrackList.size() ,1);
  for (vector<L3_Track>::iterator TrackIt = TrackList.begin();
       TrackIt != TrackList.end(); ++TrackIt) {
    L3_Track& Track = *TrackIt;
    Track.kalTransport(1344);
    rhp_hfill(MYID(8000),Track.x(),Track.y(),1);
  }
}

void L3_Monitor::put(vector<L3_SpacePoint> PointList) {
  rhp_hf1(MYID(6001), PointList.size() ,1);
  for (vector<L3_SpacePoint>::iterator p = PointList.begin();
       p != PointList.end(); ++p ) {
    int gesl = p->gesl();
    if( gesl>0 && gesl<8 ){
      int ID = 7000 + gesl;
      rhp_hfill(MYID(ID),p->x(),p->y(),1);
    }
  }
}

void L3_Monitor::put(vector<L3_Vertex> v)
{
  rhp_hf1(MYID(6003), v.size() ,1);
  for(vector<L3_Vertex>::iterator i=v.begin();i!=v.end();i++)
    {
      rhp_hf1(MYID(9000),i->x(),1);
      rhp_hf1(MYID(9001),i->y(),1);
      rhp_hf1(MYID(9002),i->z(),1);
      const L3_Target& t = L3_Geometry::ClosestTarget( i->vertex() );
      rhp_hf1(MYID(9003),t.wnr(),1);
    }
}

#endif




