// *****************************************************************************
//
// source:    L3_Trigger.C
//
// type:      source code
//
// created:   06 Dec 1999
//
// author:    Stefan Scharein, Ullrich Schwanke
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: scharein@ifh.de, schwanke@ifh.de
//
// symbol:    
//
// changes:
//
// *****************************************************************************
#include "L3_Trigger.hh"
#include <cstring>
#include <iostream>
#include "L3_Macros.hh"
#include "L3_SLP.hh"
#include "L3_Geometry.hh"
#include "L3_Parameter.hh"
#include "L3_Silicon.hh"
#include "L3_Impact.hh"
#include "L3_State.hh"

L3_Trigger::L3_Trigger() : flag(0) {  
}

L3_Trigger::L3_Trigger(const L3_Trigger& a) { *this = a; }

L3_Trigger& L3_Trigger::operator = (const L3_Trigger& a) {
  State = a.State;
  Timer = a.Timer;
#ifndef L3_NO_RHP
  Monitor = a.Monitor;
#endif
#ifdef L3_ARTE_FRAME
  ArteInterface = a.ArteInterface;
#endif
  DTLT_Table = a.DTLT_Table;
  flag = a.flag;
  return *this;
}

void L3_Trigger::init(int argc, char** argv) {  
  L3_Parameter::init(argc,argv);
  L3_Geometry::initialize();
  L3_SpacePoint::initFitMatrizen();
#ifndef L3_NO_RHP
  Monitor.init();
#endif
  L3_SLP::init();
}

void L3_Trigger::process() {
  //Timer.reset();
  L3_SLP::fill();          
  Timer.start_Trigger();
  Event.release();
  Event.fill();
  Event.reconstruct(Timer);
  trigger();
  State.update(Event);
  DTLT_Table.fill(Event, State);
  Timer.stop_Trigger();
  if(L3_Parameter::Info()) Timer.print();
  State.read(Timer);

#ifdef L3_PRINT_EVENT_SUMMARY
  if(L3_Parameter::Info()){
    printEventSummaryHeader();
    State.print();
    State.printEVHD();
    Event.print();
    Event.printRecoSummary();
    printEventSummaryFooter();
  }
#endif
#ifdef L3_ARTE_FRAME
  if(L3_Parameter::run()!=-1) ArteInterface.fill(Event);
#endif
#ifndef L3_NO_RHP
  Monitor.put(Event);
  Monitor.put(State);
  Monitor.put(Timer);
#endif  
}

void L3_Trigger::update() {
  L3_SpacePoint::setFitMatrizen();
}

void L3_Trigger::stop() {
#ifndef L3_NO_RHP
  Monitor.stop();
#endif
  if(L3_Parameter::Info()) State.printSummary();
}

L3_Event L3_Trigger::getEvent() const { return Event; }
L3_DTLT L3_Trigger::getDTLT() const { return DTLT_Table; }
const L3_Timer& L3_Trigger::getTimer() const { return Timer; }
int L3_Trigger::trig(){ return flag & 1; }

#ifdef L3_ARTE_FRAME
#include "esp.h"
int showD(){
  cout << "************ DTLT-Table **********" << endl;
  if( !ArteTable<DTLT>::booked() ){
    cout << "*           not booked           *" << endl;
  } else {
    ETLT d;
    char *p = (char*)&ArtePointer<DTLT>(2)->bits[0];
    int n = (ArteTable<DTLT>::nrow()-1)*sizeof(int)/sizeof(d);
    cout << n << " tracks filled in" << endl;
    for(int i=0;i<n;i++){
      ESP_get_int((int*)p,(int*)&d,sizeof(d)/sizeof(int));
      d.show();
      p += sizeof(d);
    }
  }
  cout << "**********************************" << endl;
  return 0;
}
#endif

void L3_Trigger::trigger() {
     //loop over all wires
     vector<L3_Track> tra = Event.silicon().Tracks();
     int maxCount = 0;
     for(int w=0;w<L3_Geometry::nTarget();w++){
       int count = 0;
       const L3_Target& target = L3_Geometry::getTarget(w);
       //loop over all tracks 
       for(vector<L3_Track>::iterator i=tra.begin();i!=tra.end();i++){
	 L3_Impact imp1d(*i,target);
	 if( imp1d.mag() > 0.1 ) continue; //ignore tracks with impact paramters above 1 mmm
	 if( imp1d.mag()/sqrt(imp1d.variance())<=2.5 ) continue; //at least 2.5 sigmas
	 int Nprim = Event.origin().nPrimVerticesOnWire(w);
	 //find closest vertex (if any)
	 int vertex = -1;
	 L3_Impact imp2d(-1.0,-1.0);
	 {
	   for(int k=0;k<Nprim;k++){
	     const L3_Vertex& prim = Event.origin().PrimVerticesOnWire(w,k);
	     L3_Impact h(*i,prim,&target);
	     if( vertex<0 || h.mag()<imp2d.mag() ){
	       vertex=k;
	       imp2d = h;
	     }
	   }
	 }
	 if( imp2d.mag() > 0.15 ) continue; //ignore tracks with impact parameters to primvertex > 1.5 mm 
#if 0
	 if( vertex>=0 ){
	   L3_Vertex& prim = Event.origin().PrimVerticesOnWire(w,vertex);
	   Hep3Vector primPos = prim->Vertex();
	   float z1;
	   float d = i->flightDiff(primPos,&z1);
	   if( z1<=0.0 ) continue; //ignore tracks for which the meeting point with the
	   //anticipated B flight direction is below the primary vertex
	   if( d>=1.0 ) continue; //ignore tracks with large distance 
	 }
#endif
	 count++;
#ifdef L3_ARTE_FRAME
	 ETLT d;
	 d.wnr = w+1;
	 d.imp1d = imp1d.mag();
	 d.simp1d = sqrt(imp1d.variance());
	 d.imp2d = imp2d.mag();
	 if( d.imp2d<0.0 ) d.simp2d = d.imp2d; else d.simp2d = sqrt(imp2d.variance());
	 d.chi2ndof = i->chisqr()/float(i->numHits()-4);
	 d.fom = 0.0;
	 d.theta = acos(1.0/sqrt(1.0+pow(i->mx(),2)+pow(i->my(),2)));
	 d.flag = 0;
	 L3_Arte::fillDTLT((int*)&d,sizeof(d)/sizeof(int));
#endif
       }
       if( count>maxCount ) count=maxCount;
     }
     if( maxCount>2 ) 
       flag = 1; //demand at least two tracks passing trigger criteria
     else
       flag = 0;
#ifdef L3_ARTE_FRAME
     if(L3_Parameter::Info()) showD();
#endif
}

void L3_Trigger::printEventSummaryHeader() const {
  cout << "*******************************************************************"
       << endl;
  cout << "*                 L3_Trigger: EventSummary                        *"
       << endl;
  cout << "*******************************************************************"
       << endl;
}

void L3_Trigger::printEventSummaryFooter() const {
  cout << "*******************************************************************"
       << endl;
  cout << "*                    End of EventSummary                          *"
       << endl;
  cout << "*******************************************************************"
       << endl;
}




