#ifndef _L3_TRACK_HH_
#define _L3_TRACK_HH_
// *****************************************************************************
//
// source:    L3_Track.hh
//
// type:      include file
//
// created:   06 Dec 1999
//
// authors:   Stefan Scharein, Ullrich Schwanke
//            HERA-B Collaboration
//            DESY Zeuthen
//            Platanenallee 6
//            D-15738 Zeuthen
//            Germany
//            email: scharein@ifh.de, schwanke@ifh.de
//
// changes:
//
// *****************************************************************************

#include "L3_Defines.hh"
#include <iostream>
#include <iomanip>
#include <cmath>
#include <vector>
#include <functional>
#include <algorithm>

#include <CLHEP/Vector/ThreeVector.h>
#include <CLHEP/Matrix/Matrix.h>
#include <CLHEP/Matrix/Vector.h>

#include "L3_Hitb.hh"
#include "L3_SpacePoint.hh"
#include "L3_Window.hh"
#include "L3_Parameter.hh"
#include "L3_Geometry.hh"
#include "L3_Mathlib.hh"
#include "L3_CovMatrix.hh"

//#define L3_DEBUG_TRACK

class L3_Track {
public:
  inline L3_Track();
  inline L3_Track(const L3_Track&);
  inline L3_Track(const int&, const L3_SpacePoint&);
  inline L3_Track(const int& nr, const float& x, const float& y, const float& mx, const float& my,
		  const float& refz);
  L3_Track(const int& nr, const float& x, const float& y, const float& mx, const float& my,
		  const float& refz,vector<L3_Hitb> &hits,float p);
  L3_Track(const int& nr, const float& x, const float& y, const float& mx, const float& my,
		  const float& refz,vector<L3_SpacePoint*> &points,float p);
  inline L3_Track& operator = (const L3_Track&);
  inline void add(const L3_SpacePoint&);
  inline void add(const vector<L3_SpacePoint>&);
  inline bool operator < (const L3_Track&) const;
  inline int operator == (const L3_Track&) const;
  inline int operator != (const L3_Track&) const;
  inline ostream& operator << (const L3_Track&);
  inline int nr() const;
  inline void setnr(int val);
  inline float x() const;
  inline float y() const;
  inline float mx() const;
  inline float my() const;
  inline float refz() const;
  inline float chisqr() const;
  inline float chisqrCut() const;
  inline float prob() const;
  inline float Cx() const;
  inline float Cmx() const;
  inline float Cmxx() const;
  inline float Cy() const;
  inline float Cmy() const;
  inline float Cmyy() const;
  inline double Cov(const int&, const int&) const;
  inline L3_CovMatrix<4> getCov() const;
  //----------------------------------------
  inline void SetCov(int i,int j,double val);
  inline void Setx(float x);
  inline void Sety(float y);
  inline void Setmx(float mx);
  inline void Setmy(float my);
  inline void Setrefz(float z);
  //----------------------------------------
  inline void fit();
  inline void fit(const float&);
  inline void kalTransport(const double&);
  void refit(float p=-1); //momentum <= 0 means 'unknown', [p] = GeV!!
  void fullrefit(float p=-1); //momentum <= 0 means 'unknown', [p] = GeV!!
  inline int numPunkte() const;
  int numPunkteSL(int sl) const;
  inline int numHits() const;
  inline int numHitsX() const;
  inline int numHitsY() const;
  inline L3_Window window(const float&);
  L3_Window window2(const float&);
  inline L3_Window window();
  inline vector<L3_SpacePoint> getPunktList() const;
  inline vector<L3_Hitb> HList() const;

  inline int getCommonHits(const L3_Track &a) const;
  inline Hep3Vector MinDistanceToPoint(const Hep3Vector &v) const;
  inline Hep3Vector TrackDistance( const L3_Track &a ) const;
  inline Hep3Vector TrackDistance2( const L3_Track &a ) const;
  inline Hep3Vector MinDistanceToTarget( const L3_Target &t ) const;
  inline Hep3Vector Vector() const;
  inline Hep3Vector UnitVector() const;
  inline float AngleWithTrack( const L3_Track &a ) const;
  inline float AngleWithVector( const Hep3Vector &v ) const;

  inline double scaleCovMatrix();
  inline float eta() const;
  inline float phi() const;
  inline int getFlag() const;
  inline void setFlag(int flag);
  inline int startingSL() const;
  inline HepVector Parameter() const;
  float lsmfit(float refz);
  float flightDiff(Hep3Vector &pv,float *z1);

private:
  typedef const L3_SpacePoint* SpacePoint;

  void parTransport(const float&);
  inline void calcSpur(const float&);
  inline void ncalcSpur(const float&);
  inline void mscalcSpur(const float&);
  void kalMultScat(float ms,float p=-1.0); //momentum <= 0 means 'unknown', [p] = GeV!!
  void kalFilter(double&, const double, const double, const double);  
  int the_nr;
  int the_flag;
  float the_x;
  float the_y;
  float the_mx;
  float the_my;
  float chiSquare;
  L3_CovMatrix<4> cov;
  float the_refz;
  vector<SpacePoint> RaumPunktList;
  vector<L3_Hitb> hitlist;

  //needed for track refit
  static int Chips[7][4][2]; // [SL][SEC][chip1 or chip2]
  inline static void resetChips(){ memset(Chips,0,sizeof(Chips)); }
};

#include "L3_Track.icc"

#endif
