/**********************************************************
 Routine to fill OTDR table.

 Should be called during initialization,
 but after filling mask status bytes.
 
 A.Lanyov    September 1999
 **********************************************************/
 
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include <arte/arte.hh>
#include <arte/artefunc.h>

#include "wax/DecodeManager.hh"
#include "wax/GedeManager.hh"

//#define DEBUG
//#define PRINT

extern "C" int otr_gede_status(int igede_, int mono, int wire) ;
extern "C" int otr_gede_status_(int *igede, int *mono, int *iwire) ;

extern "C" {

int otdr_declare(int igede, int sub, float xmin, float xmax){
  /********************************************************************
    Create a OTDR table.

    If previous OTDR table and this one can be merged together,
    New table is not created, but max of previous OTDR table is modified.

    ********************************************************************/
  ArtePointer<GEDE> gede(igede);
  int nlast = ArteTable<OTDR>::size();
#ifdef PRINT
      char hwc[13];
      strncpy(hwc,gede->hwc,12);
      hwc[12] = 0;
#endif
  
  if (nlast > 0) {
    ArtePointer<OTDR> otdr_last(ArtePointer<OTDR>::nrow());
    if (igede == (int)otdr_last->gede.index()
	&& sub == otdr_last->sub
	&& (xmin - otdr_last->xmax) > 0.5*gede->pit) {
      otdr_last->xmax = xmax;
#ifdef PRINT
      printf("Changed OTDR GEDE %4d %s, sub %1d,%8.2f, %8.2f\n",
	     igede,hwc,sub,xmin,xmax);
#endif
      return 0;
    };
  };

  mmbook_(ArtePointer<OTDR>::getLink().getRef(), 1);
  ArtePointer<OTDR> otdr(ArtePointer<OTDR>::nrow()); 
  
  strncpy(otdr->hwc,gede->hwc,12); /* hardware code for the given detector */
  otdr->xmin = xmin;               /* xmin -  lower boundary in x */
  otdr->xmax = xmax;               /* xmax -  upper boundary in x */
  otdr->sub = sub;                 /* sublayer number: 1 or 3 */
  otdr->gede = igede;              /* gede identifier */
#ifdef PRINT
  printf("Created OTDR GEDE %4d %s, sub %1d,%8.2f, %8.2f\n",
	 igede,hwc,sub,xmin,xmax);
#endif

  return 0;
}

 
int otdr_init(){
  /********************************************************************
    Create OTDR tables.
 
    ********************************************************************/
 
  ArteTable<OTDR>::clear();
 
  for ( ArteTableIterator<GEDE> gedeIt; gedeIt.valid(); gedeIt.next() ) {
    
    if (gedeIt->cmp != 2 || gedeIt->sec == 0) {
      continue;
    };
 
    int igede = gedeIt.index();
    char hwc[13];
    strncpy(hwc,gedeIt->hwc,12);
    hwc[12] = '\0';
 
    // limit 1.33*pitch corresponds to minimal DR size of 3 cells
    float limit = 1.33 * gedeIt->pit;
    
    for (int slay=0; slay < (gedeIt->nsl)/2; slay++) {
      int mono;
      float x1 = -1000.;
      float x2 = -1000.;
      float xgood = -1000.;

      int nwimax = 0;
      for (mono=2*slay; mono <= 2*slay+1; mono++) {
        if (nwimax<gedeIt->nwi[mono])
          nwimax = gedeIt->nwi[mono];
      };
      
      for (int wire=0; wire < nwimax; wire++) {
 
        for (gedeIt->xfw[2*slay]<gedeIt->xfw[2*slay+1] ?
	       mono = 2*slay    : mono = 2*slay+1;
             gedeIt->xfw[2*slay]<gedeIt->xfw[2*slay+1] ?
	       mono <= 2*slay+1 : mono >= 2*slay;
             gedeIt->xfw[2*slay]<gedeIt->xfw[2*slay+1] ?
	       mono++           : mono--) {
          
          if (wire >= gedeIt->nwi[mono])
            continue;
          
          int status = otr_gede_status(igede, mono, wire);
 
          //      if (status != -1) {
          //        printf("%12s, igede %4d, mono %1d, wire %2d, status %d \n",
          //               hwc,igede,mono,wire,status);
 
 
          x1 = gedeIt->xfw[mono] + gedeIt->pit * (wire-1./3.);
          x2 = gedeIt->xfw[mono] + gedeIt->pit * (wire+1./3.);
 
          if (xgood < -1.e9)
            xgood = x1;                    // first wire
 
          if (status == 0 ) {              // good wire
            if (x1-xgood > limit)
              otdr_declare(igede, 2*slay+1, xgood, x1);
            xgood = x2;
          };
        };
      };
 
      if (x1-xgood > limit)
        otdr_declare(igede, 2*slay+1, xgood, x2);
    };
#ifdef PRINT
    printf("\n");
#endif
  };
  
  printf("* otdr_init: %d OTDR tables were booked.       *\n",
         ArteTable<OTDR>::size());
  return 0;
} 
 
}
