#include "wax/Event.hh"
#include "wax/Hit.hh"
#include "wax/HITB.hh"

#include <arte/artefunc.h>

//__________________________________________________________________
// waxEvent
//
// Container for:
// * Hits
// Functionality:
// * Read new hits
// * Append to HITB
//
// Author: Wouter Hulsbergen
//__________________________________________________________________

// static const unsigned char fgMinTdcCount[2] = { 50,20 } ;
// static const unsigned char fgMaxTdcCount[2] = { 255,255 } ;

waxEvent::waxEvent() : fId(-1), fHits() {}

waxEvent::~waxEvent() {}

void
waxEvent::NextMCEvent()
{
  // Processing monte carlo data. For the moment: reset the decoding
  // structures and give all channels that have entry in HITB bank a
  // `valid' drifttime.
#ifdef __WAX_DEBUG__
  std::cout << "* waxEvent::NextMCEvent()" << std::endl ;
#endif
  
  // Reset tdc counts
  waxDecodeManager::GetMe()->ResetTdcCounts() ;
  
  // Set tdc coutns of channels that are hit to 100
  for(ArteTable<HITB>::iterator hitbIt = ArteTable<HITB>::begin() ;
      hitbIt != ArteTable<HITB>::end() ; hitbIt++) 
    if(hitbIt->gede->cmp==2) {
      waxHITB otrhitb = waxHITB(hitbIt) ;
      const waxDecodeWire* wire = otrhitb.GetWire() ;
      if(wire && wire->IsConnected()) 
	(const_cast<waxDecodeChannel*>(wire->GetChannel()))->SetTdcCount(100) ;
    }
}

void
waxEvent::NextMDSTEvent()
{
  // On some mini dsts the hits are already filled. To allow for
  // calibration updates we recalculate the position and the drift
  // radius. Furthermore, the pointers in the gede bank must be set.
#ifdef __WAX_DEBUG__
  std::cout << "* Re-processing mini-dst hitbs." << std::endl ;
#endif
  
  // Reset tdc counts
  waxDecodeManager::GetMe()->ResetTdcCounts() ;
  
  // re-process hitbs 
  for(ArteTable<HITB>::iterator hitbIt = ArteTable<HITB>::begin() ;
      hitbIt != ArteTable<HITB>::end() ; hitbIt++) 
    if(hitbIt->gede->cmp==2) {
      waxHITB otrhitb = waxHITB(hitbIt) ;
      const waxDecodeChannel* channel = otrhitb.GetChannel() ;
      const_cast<waxDecodeChannel*>(channel)->
	SetTdcCount(int(hitbIt->y)) ;
      waxHit otrhit(channel) ;
      
      // You can append the hit to the old place
      otrhit.AppendHITB(hitbIt) ;
      // or set only those parts that are supposed to change
      // hitbIt->x = otrhit->GetRawX() ;
      // hitbIt->z = otrhit->GetRawZ() ;
      // hitbIt->q = ... ;
      // hitbIt->r = ... ;
    }
  waxGedeManager::GetMe()->SetHITBPointers() ;
}

void
waxEvent::NextDOTREvent()
{
  // Processing raw data
  // cout << "waxEvent::Next(): Processing data event." << endl ;
  
  fHits.erase( fHits.begin(), fHits.end() );
  waxDecodeManager* dm = waxDecodeManager::GetMe() ;
  
  // Retrieve new event
  if(dm->FillNewEvent()==0) {
    // Copy hits from Decode structures. The wires are already in the
    // correct order for filling in the HITB bank.

    const waxDecodeChannel* chanIt ;
    for(waxDecodeManager::WireContainer::const_iterator 
	  wireIt = dm->GetWires().begin() ;
	wireIt != dm->GetWires().end() ; ++wireIt) {
      if(wireIt->IsConnected()) {
	chanIt = wireIt->GetChannel() ;
	if( (chanIt->GetStatus()==0) &&
	    (chanIt->GetTdcCount()>20)  &&
	    (chanIt->GetTdcCount()<0xfe) ) {
	  fHits.push_back(waxHit(chanIt)) ;
	}
      }
    }
    
    // Allocate space in the HITB bank and append hits
    ArtePointer<HITB> hitbIt = 
      ArteTable<HITB>::allocate(fHits.size()) ;
    for(HitContainer::iterator hitIt = fHits.begin() ;
	hitIt != fHits.end() ; hitIt++, hitbIt++)
      (*hitIt).AppendHITB(hitbIt) ; 
    
    // Set pointers in GEDE bank
    waxGedeManager::GetMe()->SetHITBPointers() ;
  }
}

void
waxEvent::Next()
{
  // Process next event. Depending on which tables are booked, the
  // event is processed as being monte carlo, mini-dst or raw data.
  int id = ArteTable<EVHD>::begin()->evt ;
  if(id == fId) {
    std::cout << "waxEvent::Next(...): Event already processed. Skipping." << std::endl ;
  } else {
    fId = id ;
    if( ArteTable<MCIN>::booked())        NextMCEvent() ;
    else if( !ArteTable<DOTR>::booked() ) NextMDSTEvent() ;
    else if( ArteTable<DOTR>::booked() )  NextDOTREvent() ;
  }
}

void
waxEvent::NextTATREvent(ArtePointer<TATR> tatr)
{
  // Fills HITBs corresponding to one TATR entry for the OTR
  // alignment and calibration.
  
  waxDecodeManager* dm = waxDecodeManager::GetMe() ;

  // Reset tdc counts ?
  // (not really necessary and maybe a bit slow !)
  dm->ResetTdcCounts() ;
  
  // Clear hitb bank and reserve space
  ArteTable<HITB>::clear() ;
  ArtePointer<HITB> hitbIt = 
    ArteTable<HITB>::allocate(tatr->tahbl - tatr->tahb1 + 1) ;
  
  // Loop over tahb entries
  for(ArteTable<TAHB>::iterator tahbIt = tatr->tahb1 ;
      tahbIt != ArteTableIterator<TAHB>(tatr->tahbl) + 1 ;
      tahbIt++, hitbIt++) {
    // Get channel from decode structures
    const waxDecodeChannel* chanIt = 
      dm->GetAsd(tahbIt->channel/16)->GetChannel(tahbIt->channel%16) ;

    // Set tdc value
    const_cast<waxDecodeChannel*>(chanIt)->SetTdcCount(tahbIt->dt) ;

    // Create hit and append
    (waxHit(chanIt)).AppendHITB(hitbIt) ;
  } 
  // Set pointers in GEDE bank
  waxGedeManager::GetMe()->SetHITBPointers() ;
}
